"""The cimsparql.model module contains the base class CimModel for both graphdb.GraphDBClient with
function get_table as well as a set of predefined CIM queries.
"""

import re
from abc import ABC, abstractmethod
from datetime import datetime
from functools import cached_property
from typing import Dict, Iterable, List, Optional, Tuple, Union

import numpy as np
import pandas as pd

from cimsparql import line_queries, queries
from cimsparql import query_support as sup
from cimsparql import ssh_queries, sv_queries, tp_queries
from cimsparql.constants import union_split
from cimsparql.enums import (
    ConverterTypes,
    GeneratorTypes,
    Impedance,
    LoadTypes,
    Power,
    Rates,
    SyncVars,
    TapChangerObjects,
)
from cimsparql.type_mapper import TypeMapper
from cimsparql.typehints import Region
from cimsparql.url import Prefix


class Model(Prefix, ABC):
    def __init__(self, mapper: Optional[TypeMapper], *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._mapper = mapper

    @abstractmethod
    def _get_table(self, *args, **kwargs) -> Tuple[pd.DataFrame, Dict[str, str]]:
        pass

    @abstractmethod
    def _col_map(self) -> Dict[str, str]:
        pass

    @cached_property
    def mapper(self) -> Optional[TypeMapper]:
        """Mapper used to convert str → type described by ontology in Graphdb

        Getter:
           Returns a mapper that can be used by self or other instances
        Setter:
           Sets mapper for self. Query GraphDB if not provided (None)
        """
        if self._mapper is None and "rdfs" in self.prefixes:
            return TypeMapper(self)
        return self._mapper

    def _query_with_header(self, query: str, limit: Optional[int] = None) -> str:
        query = "\n".join([self.header_str(query), query])
        if limit is not None:
            query += f" limit {limit}"
        return query

    @property
    def empty(self) -> bool:
        """Identify empty GraphDB repo"""
        try:
            self._get_table("SELECT * \n WHERE { ?s ?p ?o }", limit=1)
            return False
        except IndexError:
            return True

    @staticmethod
    def _assign_column_types(
        result: pd.DataFrame, columns: Dict[str, Union[bool, str, float, int]]
    ) -> None:
        for column, column_type in columns.items():
            if column_type is str:
                continue
            if column_type is bool:
                result[column] = result[column].str.contains("true", flags=re.IGNORECASE)
            else:
                result.loc[result[column] == "None", column] = ""
                result[column] = pd.to_numeric(result[column], errors="coerce").astype(column_type)

    @classmethod
    def col_map(cls, data_row, columns: Dict[str, str]) -> Dict[str, str]:
        columns = columns or {}
        col_map = cls._col_map(data_row)
        col_map.update(columns)
        return col_map

    def get_table(
        self,
        query: str,
        index: Optional[str] = None,
        limit: Optional[int] = None,
        map_data_types: bool = True,
        columns: Optional[Dict] = None,
    ) -> pd.DataFrame:
        """Gets given table from the configured database.

        Args:
           query: to sparql server
           index: column name to use as index
           limit: limit number of resulting rows
           map_data_types: gets datatypes from the configured graphdb & maps the types in the result
                  to correct python types
           custom_maps: dictionary of 'sparql_datatype': function to apply on columns with that
                  type. Overwrites sparql map for the types specified.
           columns: dictionary of 'column_name': function,
                  uses pandas astype on the column, or applies function.
                  Sparql map overwrites columns when available

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> query = 'select * where { ?subject ?predicate ?object }'
           >>> gdbc.get_table(query, limit=10)
        """
        try:
            result, data_row = self._get_table(query, limit)
        except IndexError:
            return pd.DataFrame([])

        if map_data_types and self.mapper is not None:
            col_map = self.col_map(data_row, columns)
            result = self.mapper.map_data_types(result, col_map)

        if index and not result.empty:
            result.set_index(index, inplace=True)
        return result

    @property
    def map_data_types(self) -> bool:
        return "cim" in self.prefixes and self.mapper.have_cim_version(self.prefixes["cim"])

    @classmethod
    def _manual_convert_types(
        cls, df: pd.DataFrame, columns: Optional[Dict], index: Optional[str]
    ) -> pd.DataFrame:
        if columns is None:
            columns = {}
        reset_index = index is not None
        if reset_index:
            df.reset_index(inplace=True)
        df = df.astype(str)
        cls._assign_column_types(df, columns)
        if reset_index:
            df.set_index(index, inplace=True)
        return df

    def _get_table_and_convert(
        self,
        query: str,
        limit: Optional[int] = None,
        index: Optional[str] = None,
        columns: Optional[Dict[str, str]] = None,
    ) -> pd.DataFrame:
        result = self.get_table(query, index, limit, map_data_types=True, columns=columns)

        if not self.map_data_types and len(result) > 0:
            result = self._manual_convert_types(result, columns, index)

        return result


class CimModel(Model):
    """Used to query with sparql queries (typically CIM)"""

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._date_version = None

    @property
    def date_version(self) -> datetime:
        """Activation date for this repository"""
        if self._date_version:
            return self._date_version
        repository_date = self._get_table_and_convert(queries.version_date())
        self._date_version = repository_date["activationDate"].values[0]
        if isinstance(self._date_version, np.datetime64):
            self._date_version = self._date_version.astype("<M8[s]").astype(datetime)
        return self._date_version

    def full_model(self, dry_run: bool = False) -> Union[pd.DataFrame, str]:
        query = queries.full_model()
        if dry_run:
            return self._query_with_header(query)
        return self._get_table_and_convert(query)

    def bus_data(
        self,
        region: Region = None,
        sub_region: bool = False,
        limit: Optional[int] = None,
        dry_run: bool = False,
        with_dummy_buses: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query name of topological nodes (TP query).

        Args:
           region: Limit to region (use None to get all)
           sub_region: True - assume sub regions, False - assume region
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = queries.bus_data(region, sub_region)
        if with_dummy_buses:
            dummy_bus_query = queries.three_winding_dummy_bus(region, sub_region)
            combined = sup.combine_statements(query, dummy_bus_query, split=union_split)
            query = sup.combine_statements(
                sup.select_statement(["?mrid", "?name", "?busname", "?un", "?area", "?bidzone"]),
                f"where {{{{{combined}}}}}",
            )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def phase_tap_changers(
        self,
        region: Region = None,
        sub_region: bool = False,
        with_tap_changer_values: bool = True,
        impedance: Iterable[Impedance] = tuple(Impedance),
        tap_changer_objects: Iterable[TapChangerObjects] = tuple(TapChangerObjects),
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Get list of phase tap changers"""
        query = queries.phase_tap_changer_query(
            region, sub_region, with_tap_changer_values, impedance, tap_changer_objects
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def loads(
        self,
        load_type: LoadTypes,
        load_vars: Optional[Iterable[Power]] = None,
        region: Region = None,
        sub_region: bool = False,
        connectivity: Optional[Optional[str]] = None,
        nodes: Optional[Optional[str]] = None,
        station_group_optional: bool = True,
        station_group: bool = False,
        with_sequence_number: bool = False,
        network_analysis: bool = True,
        limit: Optional[int] = None,
        dry_run: bool = False,
        ssh_graph: Optional[str] = None,
    ) -> Union[pd.DataFrame, str]:
        """Query load data

        Args:
           load_type: List of load types. Allowed: "ConformLoad", "NonConformLoad", "EnergyConsumer"
           load_vars: List of additional load vars to return. Possible are 'p' and/or 'q'.
           region: Limit to region
           sub_region: Assume region is a sub_region
           connectivity: Include connectivity mrids
           station_group: return station group mrid (if any)
           with_sequence_number: Include the sequence numbers in output
           network_analysis: Include only network analysis enabled components
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Returns:
           DataFrame: with mrid as index and columns ['terminal_mrid', 'bid_marked_code', 'p', 'q',
           'station_group']

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.loads(load_type=['ConformLoad', 'NonConformLoad'])
        """
        query = queries.load_query(
            load_type,
            load_vars,
            region,
            sub_region,
            connectivity,
            nodes,
            station_group_optional,
            with_sequence_number,
            network_analysis,
            station_group,
            self.cim_version,
            ssh_graph,
        )
        if dry_run:
            return self._query_with_header(query, limit)

        return self._get_table_and_convert(query, limit, index="mrid")

    def wind_generating_units(
        self,
        limit: Optional[int] = None,
        network_analysis: bool = True,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query wind generating units

        Args:
           network_analysis: Include only network analysis enabled components
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Returns:.
           wind_generating_units: with mrid as index and columns ['station_group', 'market_code',
           'maxP', 'allocationMax', 'allocationWeight', 'minP', 'name', 'power_plant_mrid']

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.wind_generating_units(limit=10)

        """
        query = queries.wind_generating_unit_query(network_analysis)
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def synchronous_machines(
        self,
        sync_vars: Iterable[SyncVars] = tuple(SyncVars),
        region: Region = None,
        sub_region: bool = False,
        connectivity: Optional[str] = None,
        nodes: Optional[str] = None,
        station_group_optional: bool = True,
        with_sequence_number: bool = False,
        network_analysis: bool = True,
        u_groups: bool = False,
        limit: Optional[int] = None,
        dry_run: bool = False,
        ssh_graph: Optional[str] = None,
    ) -> Union[pd.DataFrame, str]:
        """Query synchronous machines

        Args:
           synchronous_vars: additional vars to include in output
           region: Limit to region
           sub_region: Assume region is a sub_region
           connectivity: Include connectivity mrids
           station_group_optional: Assume station group is optional
           with_sequence_number: add this numbers
           network_analysis: query SN:Equipment.networkAnalysisEnable
           u_groups: Filter out station groups where name starts with 'U-'
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.synchronous_machines(limit=10)
        """
        query = queries.synchronous_machines_query(
            sync_vars,
            region,
            sub_region,
            connectivity,
            nodes,
            station_group_optional,
            self.cim_version,
            with_sequence_number,
            network_analysis,
            u_groups,
            ssh_graph,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def connections(
        self,
        rdf_types: Union[str, Iterable[str]] = ("cim:Breaker", "cim:Disconnector"),
        region: Region = None,
        sub_region: bool = False,
        connectivity: Optional[str] = None,
        nodes: Optional[str] = None,
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query connectors

        Args:
           rdf_types: Only cim:breaker and cim:Disconnector allowed
        Returns:

        Example:
           region: Limit to region
           sub_region: Assume region is a sub_region
           limit: return first 'limit' number of rows
           connectivity: Include connectivity mrids
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.connections(limit=10)
        """
        query = queries.connection_query(
            self.cim_version, rdf_types, region, sub_region, connectivity, nodes
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def borders(
        self,
        region: Union[str, List[str]],
        sub_region: bool = False,
        nodes: Optional[str] = None,
        ignore_hvdc: bool = True,
        with_market_code: bool = False,
        market_optional: bool = False,
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Retrieve ACLineSegments where one terminal is inside and the other is outside the region

        Args:
            region: Inside area
            sub_region: regions is sub areas
            ignore_hvdc: ignore ac lines with HVDC in name
            with_marked_code: include SN:Line.marketCode
            market_optional: include lines without market code if with_marked_code set
            limit: return first 'limit' number of rows
            dry_run: return string with sql query

        """
        query = line_queries.borders_query(
            self.cim_version,
            region,
            sub_region,
            nodes,
            ignore_hvdc,
            with_market_code,
            market_optional,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def converters(
        self,
        region: Region = None,
        sub_region: bool = False,
        converter_types: Iterable[ConverterTypes] = tuple(ConverterTypes),
        nodes: Optional[str] = None,
        sequence_numbers: Optional[List[int]] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        query = queries.converters(
            region, sub_region, self.in_prefixes(converter_types), nodes, sequence_numbers
        )
        if dry_run:
            return self._query_with_header(query)
        return self._get_table_and_convert(query, index="mrid")

    def transformers_connected_to_converter(
        self,
        region: Region = None,
        sub_region: bool = False,
        converter_types: Iterable[ConverterTypes] = tuple(ConverterTypes),
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query list of transformer connected at a converter (Voltage source or DC)

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           converter_types: VoltageSource, DC (cim:VsConverter, cim:DCConverterUnit)
           dry_run: return string with sql query

        """
        query = queries.transformers_connected_to_converter(
            region, sub_region, self.in_prefixes(converter_types)
        )
        if dry_run:
            return self._query_with_header(query)
        return self._get_table_and_convert(query, index="converter_mrid")

    def ac_lines(
        self,
        region: Region = None,
        sub_region: bool = False,
        limit: Optional[int] = None,
        connectivity: Optional[str] = None,
        nodes: Optional[str] = None,
        with_loss: bool = False,
        rates: Iterable[Rates] = (Rates.Normal,),
        network_analysis: bool = True,
        with_market: bool = False,
        temperatures: Optional[List[int]] = None,
        impedance: Iterable[Impedance] = tuple(Impedance),
        length: bool = False,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query ac line segments

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           limit: return first 'limit' number of rows
           connectivity: Include connectivity mrids as column name '{connectivity}_{1|2}'
           rates: include rates in output (available: "Normal", "Warning", "Overload")
           with_market: include marked connections in output
           temperatures: include temperature correction factors in output
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.ac_lines(limit=10)
        """
        query = line_queries.ac_line_query(
            self.cim_version,
            self.ns["cim"],
            region,
            sub_region,
            connectivity,
            nodes,
            with_loss,
            rates,
            network_analysis,
            with_market,
            temperatures,
            impedance,
            length,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        ac_lines = self._get_table_and_convert(query, limit=limit)
        if temperatures is not None:
            for temperature in temperatures:
                column = f"{sup.negpos(temperature)}_{abs(temperature)}_factor"
                ac_lines.loc[ac_lines[column].isna(), column] = 1.0
        return ac_lines

    def series_compensators(
        self,
        region: Region = None,
        sub_region: bool = False,
        connectivity: Optional[str] = None,
        with_loss: bool = False,
        nodes: Optional[str] = None,
        rates: Iterable[Rates] = (Rates.Normal,),
        network_analysis: bool = True,
        with_market: bool = False,
        impedance: Iterable[Impedance] = tuple(Impedance),
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query series compensators

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           limit: return first 'limit' number of rows
           connectivity: Include connectivity mrids
           network_analysis: query SN:Equipment.networkAnalysisEnable
           with_market: include marked connections in output
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.series_compensators(limit=10)
        """
        query = line_queries.series_compensator_query(
            self.cim_version,
            region,
            sub_region,
            connectivity,
            with_loss,
            nodes,
            rates,
            network_analysis,
            with_market,
            impedance,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit=limit)

    def transformers(
        self,
        region: Region = None,
        sub_region: bool = False,
        connectivity: Optional[str] = None,
        rates: Iterable[Rates] = (Rates.Normal,),
        network_analysis: bool = True,
        with_market: bool = False,
        impedance: Iterable[Impedance] = tuple(Impedance),
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query transformer windings

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           connectivity: Include connectivity mrids
           rates: include rates in output (available: "Normal", "Warning", "Overload")
           network_analysis: query SN:Equipment.networkAnalysisEnable
           with_market: include marked connections in output
           impedance: values returned
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.transformers(limit=10)
        """
        query = queries.transformer_query(
            region, sub_region, connectivity, rates, network_analysis, with_market, impedance
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit=limit)

    def two_winding_transformers(
        self,
        region: Region = None,
        sub_region: bool = False,
        rates: Iterable[Rates] = (Rates.Normal,),
        network_analysis: bool = True,
        with_market: bool = False,
        impedance: Iterable[Impedance] = tuple(Impedance),
        p_mrid: bool = False,
        nodes: Optional[str] = None,
        with_loss: bool = False,
        name: str = "?name",
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query two-winding transformer

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_region
           rates: include rates in output (available: "Normal", "Warning", "Overload")
           network_analysis: query SN:Equipment.networkAnalysisEnable
           with_market: include marked connections in output
           impedance: values returned
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.two_winding_transformers(limit=10)
        """
        query = queries.two_winding_transformer_query(
            region,
            sub_region,
            rates,
            network_analysis,
            with_market,
            p_mrid,
            nodes,
            with_loss,
            name,
            impedance,
            self.cim_version,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit=limit)

    def three_winding_transformers(
        self,
        region: Region = None,
        sub_region: bool = False,
        rates: Iterable[Rates] = (Rates.Normal,),
        network_analysis: bool = True,
        with_market: bool = False,
        impedance: Iterable[Impedance] = tuple(Impedance),
        p_mrid: bool = False,
        nodes: Optional[str] = None,
        with_loss: bool = False,
        name: str = "?name",
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query three-winding transformer. Return as three two-winding transformers.

        Args:
           region: Limit to region
           sub_region: Assume region is a sub_regionb
           rates: include rates in output (available: "Normal", "Warning", "Overload")
           network_analysis: query SN:Equipment.networkAnalysisEnable
           with_market: include marked connections in output
           impedance: values returned
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.two_winding_transformers(limit=10)
        """
        query = queries.three_winding_transformer_query(
            region,
            sub_region,
            rates,
            network_analysis,
            with_market,
            p_mrid,
            nodes,
            with_loss,
            name,
            impedance,
            self.cim_version,
        )
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit=limit)

    def disconnected(
        self, index: Optional[str] = None, limit: Optional[int] = None, dry_run: bool = False
    ) -> Union[pd.DataFrame, str]:
        """Query disconneced status from ssh profile (not available in GraphDB)

        Args:
           index: Column to use as index
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = ssh_queries.disconnected(self.cim_version)
        if dry_run:
            return self._query_with_header(query, limit)
        return self.get_table(query, index=index, limit=limit)

    def ssh_synchronous_machines(
        self, limit: Optional[int] = None, dry_run: bool = False
    ) -> Union[pd.DataFrame, str]:
        """Query synchronous machines from ssh profile (not available in GraphDB)

        Args:
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = ssh_queries.synchronous_machines()
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def ssh_load(
        self,
        rdf_types: Optional[List[str]] = None,
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query load data from ssh profile (not available in GraphDB)

        Args:
           rdf_types: allowed ["cim:ConformLoad", "cim:NonConformLoad"]
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        """
        if rdf_types is None:
            rdf_types = ["cim:ConformLoad", "cim:NonConformLoad"]
        query = ssh_queries.load(rdf_types)
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def ssh_generating_unit(
        self,
        rdf_types: Optional[List[str]] = None,
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query generating units from ssh profile (not available in GraphDB)

        Args:
           rdf_types: allowed
               ["cim:HydroGeneratingUnit", "cim:ThermalGeneratingUnit", "cim:WindGeneratingUnit"]
           limit: return first 'limit' number of rows
           dry_run: return string with sql query

        """
        if rdf_types is None:
            rdf_types = [f"cim:{unit}GeneratingUnit" for unit in GeneratorTypes]
        query = ssh_queries.generating_unit(rdf_types)
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def terminal(
        self, limit: Optional[int] = None, dry_run: bool = False
    ) -> Union[pd.DataFrame, str]:
        """Query terminals from tp profile (not available in GraphDB)

        Args:
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = tp_queries.terminal(self.cim_version)
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def topological_node(
        self, limit: Optional[int] = None, dry_run: bool = False
    ) -> Union[pd.DataFrame, str]:
        """Query topological nodes from tp profile (not available in GraphDB)

        Args:
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = tp_queries.topological_node()
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def powerflow(
        self,
        power: Tuple[str, ...] = ("p", "q"),
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query powerflow from sv profile (not available in GraphDB)

        Args:
           power: Allowed ['p','q']
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = sv_queries.powerflow(power)
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def voltage(
        self,
        voltage_vars: Iterable[str] = ("v", "angle"),
        limit: Optional[int] = None,
        dry_run: bool = False,
    ) -> Union[pd.DataFrame, str]:
        """Query voltage from sv profile (not available in GraphDB)

        Args:
           voltage_vars: allowed ["v", "angle"]
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = sv_queries.voltage(voltage_vars)
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    def tapstep(
        self, limit: Optional[int] = None, dry_run: bool = False
    ) -> Union[pd.DataFrame, str]:
        """Query tapstep from sv profile (not available in GraphDB)

        Args:
           limit: return first 'limit' number of rows
           dry_run: return string with sql query
        """
        query = sv_queries.tapstep()
        if dry_run:
            return self._query_with_header(query, limit)
        return self._get_table_and_convert(query, limit, index="mrid")

    @property
    def regions(self) -> pd.DataFrame:
        """Query regions

        Property:
           regions: List of regions in database

        Example:
           >>> from cimsparql.graphdb import GraphDBClient
           >>> from cimsparql.url import service
           >>> gdbc = GraphDBClient(service('LATEST'))
           >>> gdbc.regions
        """
        query = queries.regions_query()
        return self._get_table_and_convert(query, limit=None, index="mrid")



from functools import reduce
from operator import iconcat
from typing import Iterable, List, Literal, Optional, Tuple, Union

import cimsparql.query_support as sup
from cimsparql.cim import ACLINE, CNODE_CONTAINER, EQUIP_CONTAINER, ID_OBJ, SUBSTATION
from cimsparql.constants import sequence_numbers, union_split
from cimsparql.enums import Impedance, Rates
from cimsparql.typehints import Region


def _sv_terminal_injection(nr: int) -> str:
    return "\n".join(
        [
            f"?sv_t_{nr} cim:SvPowerFlow.Terminal ?_t_mrid_{nr}.",
            f"?sv_t_{nr} cim:SvPowerFlow.p ?sv_p_{nr}",
        ]
    )


def _line_query(
    cim_version: int,
    line_type: str,
    connectivity: Optional[str],
    nodes: Optional[str],
    with_loss: bool,
    rates: Iterable[Literal["Normal", "Warning", "Overload"]],
    network_analysis: bool,
    with_market: bool,
    impedance: Iterable[str],
) -> Tuple[List[str], List[str], str]:
    mrid_subject = "?_mrid"
    name = "?name"

    variables = [name, "?mrid", *sup.to_variables(impedance), "?un"]
    variables.extend(sup.sequence_variables(nodes if nodes else "t_mrid"))

    if connectivity:
        variables.extend(sup.sequence_variables(connectivity))

    impedance_properties = {z: f"?{z}" for z in impedance}
    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        sup.rdf_type_tripler(mrid_subject, line_type),
        sup.get_name(mrid_subject, name),
        sup.base_voltage(mrid_subject, "?un"),
        *sup.terminal_sequence_query(cim_version, connectivity, nodes, mrid_subject),
        *sup.predicate_list(mrid_subject, line_type, impedance_properties),
    ]

    sup.include_market(with_market, variables, where_list)

    if with_loss:
        variables.append("(?pl as ?pl_1) (?pl as ?pl_2)")
        loss_list = [_sv_terminal_injection(nr) for nr in [1, 2]]
        loss_list.append("bind((xsd:float(?sv_p_1) + xsd:float(?sv_p_2) ) / 2 as ?pl)")
        where_list.append(sup.group_query(loss_list, command="OPTIONAL"))

    if network_analysis:
        where_list.append(f"{mrid_subject} SN:Equipment.networkAnalysisEnable {network_analysis}")

    if rates:
        limit_type = "ActivePowerLimit" if line_type == ACLINE else "CurrentLimit"
        variables.extend([f"?rate{rate}" for rate in rates])
        where_rate: List[str] = reduce(
            iconcat, [sup.operational_limit(mrid_subject, rate, limit_type) for rate in rates], []
        )
        where_list.append(sup.group_query(where_rate, command="OPTIONAL"))
    return variables, where_list, mrid_subject


def ac_line_query(
    cim_version: int,
    cim: str,
    region: Region,
    sub_region: bool,
    connectivity: Optional[str],
    nodes: Optional[str],
    with_loss: bool,
    rates: Iterable[Rates],
    network_analysis: bool,
    with_market: bool,
    temperatures: Optional[List[int]],
    impedance: Iterable[Impedance],
    length: bool,
) -> str:
    variables, where_list, mrid_subject = _line_query(
        cim_version,
        ACLINE,
        connectivity,
        nodes,
        with_loss,
        rates,
        network_analysis,
        with_market,
        impedance,
    )
    if length:
        variables.append("?length")
        where_list.append(f"{mrid_subject} cim:Conductor.length ?length")

    if region:
        where_list.extend(
            [
                f"{mrid_subject} {EQUIP_CONTAINER} ?Line",
                *sup.region_query(region, sub_region, "Line"),
            ]
        )

    if temperatures:
        variables.extend(
            [
                f"?{sup.negpos(temperature)}_{abs(temperature)}_factor"
                for temperature in temperatures
            ]
        )
        where_list.append(
            sup.group_query(
                sup.temp_correction_factors(mrid_subject, cim, temperatures), command="OPTIONAL"
            )
        )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def series_compensator_query(
    cim_version: int,
    region: Region,
    sub_region: bool,
    connectivity: Optional[str],
    with_loss: bool,
    nodes: Optional[str],
    rates: Iterable[Rates],
    network_analysis: bool,
    with_market: bool,
    impedance: Iterable[Impedance],
) -> str:
    variables, where_list, mrid_subject = _line_query(
        cim_version,
        "cim:SeriesCompensator",
        connectivity,
        nodes,
        with_loss,
        rates,
        network_analysis,
        with_market,
        impedance,
    )

    if region:
        where_list += [
            f"{mrid_subject} {EQUIP_CONTAINER}/{SUBSTATION} ?Substation",
            *sup.region_query(region, sub_region, "Substation"),
        ]

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def borders_query(
    cim_version: int,
    region: Union[str, List[str]],
    sub_region: bool,
    nodes: Optional[str],
    ignore_hvdc: bool,
    with_market_code: bool,
    market_optional: bool,
) -> str:
    mrid_subject = "?_mrid"
    name = "?name"

    areas = sup.sequence_variables("area")
    variables = [name, "?mrid", *areas]
    variables.extend(sup.sequence_variables(nodes if nodes else "t_mrid"))

    border_filter = sup.border_filter(region, *areas)
    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        sup.get_name(mrid_subject, name),
        sup.rdf_type_tripler(mrid_subject, ACLINE),
        *sup.terminal_sequence_query(cim_version, "con", nodes, mrid_subject),
        sup.combine_statements(*border_filter, group=True, split=union_split),
    ]
    for nr in sequence_numbers:
        where_list.extend(
            [
                f"?con_{nr} {CNODE_CONTAINER}/{SUBSTATION}/cim:Substation.Region ?reg_{nr}",
                sup.region_name_query(f"?area_{nr}", sub_region, f"?reg_{nr}"),
            ]
        )

    if with_market_code:
        variables.append("?market_code")
        where_market = [
            f"{mrid_subject} {EQUIP_CONTAINER}/SN:Line.marketCode ?market_code",
        ]
        where_list.append(
            sup.group_query(where_market, command="OPTIONAL" if market_optional else "")
        )

    if ignore_hvdc:
        where_list.append(sup.combine_statements(f"FILTER (!regex({name}, 'HVDC'))", group=True))

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))



from typing import Iterable, List

from cimsparql.query_support import acdc_terminal, combine_statements, group_query


def disconnected(cim_version: int) -> str:
    disconnector = [
        "?mrid ?p cim:Disconnector",
        "?mrid cim:Switch.open ?status",
        'FILTER (?status = "true")',
    ]

    terminal = [
        "?mrid ?p cim:Terminal",
        f"?mrid cim:{acdc_terminal(cim_version)}.connected ?connected",
        'FILTER (?connected = "false")',
    ]

    where_list = group_query(
        [combine_statements(*comp, group=True, split=" .\n") for comp in [disconnector, terminal]],
        split=" UNION ",
    )
    return combine_statements("SELECT ?mrid", where_list)


def synchronous_machines() -> str:
    select = "SELECT ?mrid ?p ?q ?controlEnabled"
    where_list = [
        "?mrid rdf:type cim:SynchronousMachine",
        "?mrid cim:RotatingMachine.p ?p",
        "?mrid cim:RotatingMachine.q ?q",
        "?mrid cim:RegulatingCondEq.controlEnabled ?controlEnabled",
    ]
    return combine_statements(select, group_query(where_list))


def _load(rdf_type: str) -> str:
    return " .\n".join(
        [f"?mrid rdf:type {rdf_type}"] + [f"?mrid cim:EnergyConsumer.{s} ?{s}" for s in ["p", "q"]]
    )


def load(rdf_types: List[str]) -> str:
    query = "\nSELECT ?mrid ?p ?q\nWHERE {{\n"
    query += "\n} UNION {\n".join([_load(rdf_type) for rdf_type in rdf_types])
    query += "}}\n"
    return query


def _generating_unit(rdf_type: str) -> str:
    return " .\n".join(
        [f"?mrid rdf:type {rdf_type}", "?mrid cim:GeneratingUnit.normalPF ?normalPF"]
    )


def generating_unit(rdf_types: Iterable[str]) -> str:
    query = "SELECT ?mrid ?normalPF \n WHERE {{\n"
    query += "\n} UNION {\n".join([_generating_unit(rdf_type) for rdf_type in rdf_types])
    query += "}}\n"
    return query



import copy
from typing import List, Set

import networkx as nx
import numpy as np
import pandas as pd


class Islands(nx.Graph):
    def __init__(self, connections: pd.DataFrame) -> None:
        super().__init__()
        self.add_edges_from(connections.to_numpy())
        self._groups = list(nx.connected_components(self))

    def reference_nodes(self, columns: List[str] = None) -> pd.DataFrame:
        columns = ["mrid", "ref_node"] if columns is None else columns
        keys = []
        values = []
        for group in self.groups():
            ref = list(group)[0]
            keys += list(group)
            values += [ref] * len(group)
        return pd.DataFrame(np.array([keys, values]).transpose(), columns=columns).set_index("mrid")

    def groups(self) -> List[Set]:
        return copy.deepcopy(self._groups)



from __future__ import annotations

import warnings
from datetime import datetime
from decimal import Decimal
from typing import TYPE_CHECKING, Any, Callable, Dict, List, Optional, Union

import pandas as pd

from cimsparql.query_support import combine_statements, unionize

if TYPE_CHECKING:
    from cimsparql.model import CimModel

# Type-caster that can be used with pandas. It can be python types, numpy dtypes or string
# value. Examples: float, "int", int, "string"
TYPE_CASTER = Union[Callable[[Any], Any], str]
SPARQL_TYPE = str
COL_NAME = str
PREFIX = str
URI = str

as_type_able = [int, float, "string", "Int64", "Int32", "Int16"]


def to_timedelta(duration: str) -> datetime.timedelta:
    # Pandas only supports days, hours, minites, seconds
    # not year and month which can be part of
    # https://www.w3.org/TR/xmlschema11-2/#duration

    if "Y" in duration:
        raise ValueError("Cimsparql uses pandas to convert duration. Y not supported")

    return pd.to_timedelta(duration)


def identity(x):
    return x


XSD_TYPE_MAP = {
    # Primitive types (https://www.w3.org/TR/xmlschema11-2/#built-in-primitive-datatypes)
    "boolean": lambda x: x.lower() in {"true", "1"},
    "date": pd.to_datetime,
    "dateTime": pd.to_datetime,
    "decimal": Decimal,
    "double": float,
    "duration": to_timedelta,  # Require ISO 8601 format,
    "float": float,
    "integer": int,
    "time": pd.to_datetime,
}


CIM_TYPE_MAP = {
    "String": "string",
    "Integer": int,
    "Boolean": bool,
    "Float": float,
    "Date": pd.to_datetime,
}

sparql_type_map = {
    "literal": "string",
    "uri": "string",
}


def build_type_map(prefixes: Dict[PREFIX, URI]) -> Dict[SPARQL_TYPE, TYPE_CASTER]:
    short_map = {"xsd": XSD_TYPE_MAP, "cim": CIM_TYPE_MAP}

    type_map = {}
    for namespace, prim_types in short_map.items():
        if namespace not in prefixes:
            continue
        prefix = prefixes[namespace]
        new_map = {f"{prefix}#{dtype}": converter for dtype, converter in prim_types.items()}
        type_map.update(new_map)
    return type_map


class TypeMapperQueries:
    def __init__(self, prefixes: Dict[PREFIX, URI]):
        self.prefixes = prefixes

    @property
    def type_queries(self) -> List[List[str]]:
        queries = [
            ["?sparql_type rdf:type rdfs:Datatype", "?sparql_type owl:equivalentClass ?range"],
            ["?sparql_type owl:equivalentClass ?range"],
        ]

        if "cims" in self.prefixes:
            queries += self.cims_queries
        return queries

    @property
    def cims_queries(self) -> List[List[str]]:
        """
        Return queries that require that cims namespace. We must distinguish between types
        that is of type 'Primitive' and types that are of type 'CIMdatatype'. In the latter case
        we must find the type of the 'value' attribute
        """
        return [
            [
                "?datatypevalue rdfs:domain ?sparql_type",
                "?datatypevalue cims:dataType ?range",
                '?range cims:stereotype "Primitive"',
            ],
            [
                "?datatypevalue rdfs:domain ?sparql_type",
                "?datatypevalue cims:dataType ?CIMdtype",
                '?CIMdtype cims:stereotype "CIMDatatype"',
                "?CIMdtypeValue rdfs:domain ?CIMdtype",
                '?CIMdtypeValue rdfs:label "value"@en',
                "?CIMdtypeValue cims:dataType ?range",
            ],
        ]

    @property
    def query(self) -> str:
        select_query = "SELECT ?sparql_type ?range"
        grouped = [combine_statements(*g, split=".\n") for g in self.type_queries]
        union = unionize(*grouped)
        return f"{select_query}\nWHERE\n{{\n{union}\n}}"


class TypeMapper:
    def __init__(self, client: CimModel, custom_additions: Optional[Dict[str, Any]] = None) -> None:
        self.queries = TypeMapperQueries(client.prefixes)
        self.prefixes = client.prefixes
        custom_additions = custom_additions or {}
        self.prim_type_map = build_type_map(self.prefixes)
        self.map = sparql_type_map | self.get_map(client) | self.prim_type_map | custom_additions

    def have_cim_version(self, cim) -> bool:
        return cim in (val.split("#")[0] for val in self.map.keys())

    def type_map(self, df: pd.DataFrame) -> Dict[str, Any]:
        return {
            row.sparql_type: self.prim_type_map.get(row.range, "string") for row in df.itertuples()
        }

    def get_map(self, client: CimModel) -> Dict[str, Any]:
        """Reads all metadata from the sparql backend & creates a sparql-type -> python type map

        Args:
            client: initialized CimModel

        Returns:
            sparql-type -> python type map

        """
        df = client.get_table(self.queries.query, map_data_types=False)
        if df.empty:
            return {}
        return self.type_map(df)

    def get_type(self, sparql_type: str, col: str = ""):
        """Gets the python type/function to apply on columns of the sparql_type

        Args:
            sparql_type:
            missing_return: returns the identity-function if python- type/function is not found,
                else returns None

            custom_maps: dictionary on the form {'sparql_data_type': function/datatype} overwrites
                the default types gained from the graphdb. Applies the function/datatype on all
                columns in the DataFrame that are of the sparql_data_type

        Returns:
            python datatype or function to apply on DataFrame columns

        """

        try:
            return self.map[sparql_type]
        except KeyError:
            warnings.warn(f"{col}:{sparql_type} not found in the sparql -> python type map")

    def build_type_caster(
        self, col_map: Dict[COL_NAME, SPARQL_TYPE]
    ) -> Dict[COL_NAME, TYPE_CASTER]:
        """
        Construct a direct mapping from column names to a type caster from the
        """
        return {col: self.get_type(dtype, col) for col, dtype in col_map.items()}

    def map_data_types(
        self, df: pd.DataFrame, col_map: Dict[COL_NAME, SPARQL_TYPE]
    ) -> pd.DataFrame:
        """Maps the dtypes of a DataFrame to the python-corresponding types of the sparql-types from the
        source data

        Args:
            df: DataFrame with columns to be converted

        Returns:
            mapped DataFrame

        """
        type_caster = self.build_type_caster(col_map)
        df = map_base_types(df, type_caster)
        df = map_exceptions(df, type_caster)
        return df


def map_base_types(df: pd.DataFrame, type_map: Dict[COL_NAME, TYPE_CASTER]) -> pd.DataFrame:
    """Maps the datatypes in type_map which can be used with the df.astype function

    Args:
        df:

    Returns:
        mapped DataFrame

    """
    as_type_able_columns = {c for c, datatype in type_map.items() if datatype in as_type_able}
    if not df.empty:
        df = df.astype({column: type_map[column] for column in as_type_able_columns})
    return df


def map_exceptions(df: pd.DataFrame, type_map: Dict[COL_NAME, TYPE_CASTER]) -> pd.DataFrame:
    """Maps the functions/datatypes in type_map which cant be done with the df.astype function

    Args:
        df:

    Returns:
        mapped DataFrame

    """
    ex_columns = {c for c, datatype in type_map.items() if datatype not in as_type_able}
    for column in ex_columns:
        df[column] = df[column].apply(type_map[column])
    return df



import collections
import gzip
import re
from abc import ABC, abstractmethod
from functools import cached_property
from pathlib import Path
from typing import Callable, Dict, Iterable, List, Optional, Tuple, Union
from zipfile import ZipFile

import pandas as pd
import pendulum
from defusedxml.lxml import RestrictedElement, fromstring, parse
from lxml.etree import _ElementTree

ElementValues = Dict[str, Optional[Union[float, str, bool]]]


def attrib(node: Optional[RestrictedElement], key: str, prefix: str) -> Optional[str]:
    return None if node is None else re.sub("^[#]?_", "", node.attrib[f"{{{prefix}}}{key}"])


class CimXmlBase(ABC):
    _root = None

    @abstractmethod
    def _parse(self):
        pass

    @property
    def nsmap(self) -> Dict[str, str]:
        return self.root.nsmap

    @property
    def root(self) -> _ElementTree:
        if self._root is None:
            self._parse()
        return self._root

    def findall(self, path: str) -> List[RestrictedElement]:
        return self.root.findall(path, self.nsmap)

    @staticmethod
    def _attrib(
        node: RestrictedElement, element: str, nsmap: Dict[str, str], key: str = "resource"
    ) -> Optional[str]:
        return attrib(node.find(element, nsmap), key, nsmap["rdf"])

    @classmethod
    def _element_values(
        cls,
        node: RestrictedElement,
        nsmap: Dict[str, str],
        element: str,
        mrid: str,
        variables: Iterable[str],
    ) -> ElementValues:
        vars = {var: float(node.find(f"{element}.{var}", nsmap).text) for var in variables}
        return {"mrid": cls._attrib(node, f"{element}.{mrid}", nsmap), **vars}

    @classmethod
    def _sv_power_flow_data_adder(
        cls, node: RestrictedElement, nsmap: Dict[str, str]
    ) -> ElementValues:
        return cls._element_values(node, nsmap, "cim:SvPowerFlow", "Terminal", ["p", "q"])

    @classmethod
    def _sv_voltage_data_adder(
        cls, node: RestrictedElement, nsmap: Dict[str, str]
    ) -> ElementValues:
        return cls._element_values(node, nsmap, "cim:SvVoltage", "TopologicalNode", ["v", "angle"])

    @classmethod
    def _sv_injection_adder(cls, node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return cls._element_values(
            node, nsmap, "cim:SvInjection", "TopologicalNode", ["pInjection", "qInjection"]
        )

    @staticmethod
    def _sv_status_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(
                node.find("cim:SvStatus.ConductingEquipment", nsmap), "resource", nsmap["rdf"]
            ),
            "in_service": node.find("cim:SvStatus.inService", nsmap).text == "true",
        }

    @staticmethod
    def _sv_tap_step_data_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node.find("cim:SvTapStep.TapChanger", nsmap), "resource", nsmap["rdf"]),
            "position": int(node.find("cim:SvTapStep.position", nsmap).text),
        }

    @classmethod
    def _topological_node_data_adder(
        cls, node: RestrictedElement, nsmap: Dict[str, str]
    ) -> ElementValues:
        return {
            "name": node.find("cim:IdentifiedObject.name", nsmap).text,
            "base_voltage": cls._attrib(node, "cim:TopologicalNode.BaseVoltage", nsmap),
            "connectivity_node_container": cls._attrib(
                node, "cim:TopologicalNode.ConnectivityNodeContainer", nsmap
            ),
            "mrid": attrib(node, "ID", nsmap["rdf"]),
        }

    @staticmethod
    def _terminal_data_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        rdf = nsmap["rdf"]
        connected = node.find("cim:Terminal.connected", nsmap)
        topological_node = node.find("cim:Terminal.TopologicalNode", nsmap)
        tp_node = None if topological_node is None else attrib(topological_node, "resource", rdf)
        return {
            "mrid": attrib(node, "about", rdf),
            "connected": True if connected is None else connected.text == "true",
            "tp_node": tp_node,
        }

    @staticmethod
    def _synchrounous_machine_adder(
        node: RestrictedElement, nsmap: Dict[str, str]
    ) -> ElementValues:
        control_enabled = node.find("cim:RegulatingCondEq.controlEnabled", nsmap).text == "true"
        mode = attrib(
            node.find("cim:SynchronousMachine.operatingMode", nsmap), "resource", nsmap["rdf"]
        )
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "control_enabled": control_enabled,
            "priority": int(node.find("cim:SynchronousMachine.referencePriority", nsmap).text),
            "mode": mode,
            "p": float(node.find("cim:RotatingMachine.p", nsmap).text),
            "q": float(node.find("cim:RotatingMachine.q", nsmap).text),
        }

    @classmethod
    def _generating_unit_adder(
        cls, node: RestrictedElement, nsmap: Dict[str, str]
    ) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "pf": float(node.find("cim:GeneratingUnit.normalPF", nsmap).text),
        }

    @staticmethod
    def _regulating_control_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "discrete": node.find("cim:RegulatingControl.discrete", nsmap).text == "true",
            "enabled": node.find("cim:RegulatingControl.enabled", nsmap).text == "true",
            "deadband": float(node.find("cim:RegulatingControl.targetDeadband", nsmap).text),
            "value": float(node.find("cim:RegulatingControl.targetValue", nsmap).text),
            "multiplier": attrib(
                node.find("cim:RegulatingControl.targetValueUnitMultiplier", nsmap),
                "resource",
                nsmap["rdf"],
            ),
        }

    @staticmethod
    def _tap_changer_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "enabled": node.find("cim:TapChanger.controlEnabled", nsmap).text == "true",
            "step": int(node.find("cim:TapChanger.step", nsmap).text),
        }

    @staticmethod
    def _linear_shunt_compensator_adder(
        node: RestrictedElement, nsmap: Dict[str, str]
    ) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "enabled": node.find("cim:RegulatingCondEq.controlEnabled", nsmap).text == "true",
            "sections": int(node.find("cim:ShuntCompensator.sections", nsmap).text),
        }

    @staticmethod
    def _load_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "p": float(node.find("cim:EnergyConsumer.p", nsmap).text),
            "q": float(node.find("cim:EnergyConsumer.q", nsmap).text),
        }

    @staticmethod
    def _static_var_compensator_adder(
        node: RestrictedElement, nsmap: Dict[str, str]
    ) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "enabled": node.find("cim:RegulatingCondEq.controlEnabled", nsmap).text == "true",
            "sections": float(node.find("cim:StaticVarCompensator.q", nsmap).text),
        }

    @staticmethod
    def _control_area_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "interchange": float(node.find("cim:ControlArea.netInterchange", nsmap).text),
        }

    @staticmethod
    def _converter_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "p": float(node.find("cim:ACDCConverter.p", nsmap).text),
            "q": float(node.find("cim:ACDCConverter.q", nsmap).text),
        }

    @staticmethod
    def _terminal_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "connected": node.find("cim:ACDCTerminal.connected", nsmap).text == "true",
        }

    @staticmethod
    def _switch_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "open": node.find("cim:Switch.open", nsmap).text == "true",
        }

    @staticmethod
    def _power_limit_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "limit": float(node.find("cim:ApparentPowerLimit.value", nsmap).text),
        }

    @staticmethod
    def _current_limit_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "mrid": attrib(node, "about", nsmap["rdf"]),
            "limit": float(node.find("cim:CurrentLimit.value", nsmap).text),
        }

    @staticmethod
    def _full_model_adder(node: RestrictedElement, nsmap: Dict[str, str]) -> ElementValues:
        return {
            "time": pendulum.parse(node.find("md:Model.scenarioTime", nsmap).text),
            "created": pendulum.parse(node.find("md:Model.created", nsmap).text),
            "description": node.find("md:Model.description", nsmap).text,
            "version": node.find("md:Model.version", nsmap).text,
            "profile": node.find("md:Model.profile", nsmap).text,
            "authority": node.find("md:Model.modelingAuthoritySet", nsmap).text,
        }

    def _adder(self, profile: str) -> Callable:
        try:
            return {
                "ACDCConverterDCTerminal": self._terminal_adder,
                "ApparentPowerLimit": self._power_limit_adder,
                "Breaker": self._switch_adder,
                "ConformLoad": self._load_adder,
                "ControlArea": self._control_area_adder,
                "CsConverter": self._converter_adder,
                "CurrentLimit": self._current_limit_adder,
                "DCTerminal": self._terminal_adder,
                "Disconnector": self._switch_adder,
                "FullModel": self._full_model_adder,
                "HydroGeneratingUnit": self._generating_unit_adder,
                "LinearShuntCompensator": self._linear_shunt_compensator_adder,
                "NonConformLoad": self._load_adder,
                "PhaseTapChangerLinear": self._tap_changer_adder,
                "RatioTapChanger": self._tap_changer_adder,
                "RegulatingControl": self._regulating_control_adder,
                "StaticVarCompensator": self._static_var_compensator_adder,
                "SvInjection": self._sv_injection_adder,
                "SvPowerFlow": self._sv_power_flow_data_adder,
                "SvStatus": self._sv_status_adder,
                "SvTapStep": self._sv_tap_step_data_adder,
                "SvVoltage": self._sv_voltage_data_adder,
                "SynchronousMachine": self._synchrounous_machine_adder,
                "TapChangerControl": self._regulating_control_adder,
                "Terminal": self._terminal_data_adder,
                "ThermalGeneratingUnit": self._generating_unit_adder,
                "TopologicalNode": self._topological_node_data_adder,
                "VsConverter": self._converter_adder,
                "WindGeneratingUnit": self._generating_unit_adder,
            }[profile]
        except KeyError:
            raise NotImplementedError(f"Not implememted adder for {profile}")

    def parse(self, profile: str, ns: Optional[str] = None) -> pd.DataFrame:
        """Parse SVTP xml str

        Args:
            profile: What data to extract. Can be one of:
                     'SvVoltage'|'SvTapStep'|'TopologicalNode'|'Terminal'|'SvPowerFlow'
            for SSH file acceptable values for profile:
                     'ConformLoad'|'PhaseTapChangerLinear'|'RegulatingControl'|
                     'SynchronousMachine'|'ConformLoad'|'TapChangerControl'
        """
        if ns is None:
            ns = "cim"
        ns = "md" if profile.endswith("FullModel") else ns
        data = [self._adder(profile)(node, self.nsmap) for node in self.findall(f"{ns}:{profile}")]
        return pd.DataFrame([item for item in data if item is not None]).drop_duplicates()


class CimXml(CimXmlBase):
    def __init__(self, file_path: Path) -> None:
        self._file_path = file_path

    def _parse(self) -> None:
        if self._file_path.suffix == ".zip":
            with ZipFile(self._file_path) as cimzip:
                with cimzip.open(self._file_path.stem + ".xml") as fid:
                    self._root = parse(fid).getroot()
        else:
            open_function = gzip.open if self._file_path.suffix == ".gz" else open
            with open_function(self._file_path.absolute().as_posix(), "r") as fid:
                self._root = parse(fid).getroot()


class CimXmlStr(CimXmlBase):
    def __init__(self, text: str) -> None:
        self._text = text

    def _parse(self):
        self._root = fromstring(self._text)


class SvTpCimXml:
    def __init__(self, sv_path: Path, tp_path: Path) -> None:
        self.paths = {"sv": sv_path, "tp": tp_path}
        self._parser = {profile: CimXml(path) for profile, path in self.paths.items()}

    def __str__(self) -> str:
        file_desc = ", ".join([f"{profile}: {path.stem}" for profile, path in self.paths.items()])
        return f"<SvTpCimXml object, {file_desc}>"

    @cached_property
    def voltage(self) -> pd.DataFrame:
        return self._parser["sv"].parse("SvVoltage").set_index("mrid")

    @cached_property
    def tap_steps(self) -> pd.DataFrame:
        return self._parser["sv"].parse("SvTapStep").set_index("mrid")

    @cached_property
    def bus_data(self) -> pd.DataFrame:
        return self._parser["tp"].parse("TopologicalNode").set_index("mrid")

    @cached_property
    def terminal(self) -> pd.DataFrame:
        return self._parser["tp"].parse("Terminal").set_index("mrid")

    @cached_property
    def powerflow(self) -> pd.DataFrame:
        return self._parser["sv"].parse("SvPowerFlow").set_index("mrid")

    @cached_property
    def status(self) -> pd.DataFrame:
        return self._parser["sv"].parse("SvStatus").set_index("mrid")

    @cached_property
    def injection(self) -> pd.DataFrame:
        return self._parser["sv"].parse("SvInjection").set_index("mrid")


class CimXmlProfiles(SvTpCimXml):
    def __init__(self, sv_path: Path, tp_path: Path, ssh_path: Path) -> None:

        self.paths = {"sv": sv_path, "tp": tp_path, "ssh": ssh_path}
        self._parser = {profile: CimXml(path) for profile, path in self.paths.items()}

    @cached_property
    def model(self) -> pd.Series:
        return self._parser["ssh"].parse("FullModel").squeeze("index").rename("FullModel")

    @cached_property
    def generation(self) -> pd.DataFrame:
        return self._parser["ssh"].parse("SynchronousMachine").set_index("mrid")

    def demand(
        self,
        loads: Tuple[str, ...] = ("ConformLoad", "NonConformLoad"),
    ) -> pd.DataFrame:
        return pd.concat([self._parser["ssh"].parse(load) for load in loads]).set_index("mrid")

    @cached_property
    def tap_steps(self) -> pd.Series:
        steps = self._parser["ssh"].parse("RatioTapChanger").set_index("mrid")
        return steps.loc[steps["enabled"], "step"]

    @cached_property
    def converters(
        self, converters: Tuple[str, ...] = ("VsConverter", "CsConverter")
    ) -> pd.DataFrame:
        return pd.concat(
            [self._parser["ssh"].parse(converter) for converter in converters]
        ).set_index("mrid")


def parse_cim_file(file_name: str) -> Tuple[pendulum.DateTime, str]:
    """
    Parses a cim file-name to pendulum datetime and cim type.

    Args:
        file_name: cim file-name on format cim_YYYYMMDD_HHMMSS_foo_bar_<sv|tp>.xml

    Returns: date of cim file, and cim type

    """
    splitted = file_name.split("_")
    date = pendulum.parse(" ".join(splitted[1:3]), tz="Europe/Oslo")
    file_type = splitted[-1].split(".")[0]
    return date, file_type


def find_min(
    date: pendulum.DateTime, dates: List[pendulum.DateTime]
) -> Tuple[pendulum.DateTime, List[pendulum.DateTime]]:
    """Finds the closest date to a given date in a list of dates

    Assumes that the list of dates is sorted to reduce the iterations

    Args:
        date: Date that is to be found a closest match to
        dates: A list of dates

    Returns:
        A tuple of the date in dates that is closest to the original date and a list containing the
        unchecked dates

    """
    min_date = None
    dist = None
    for _i, d in enumerate(dates):
        dist_ = date.diff(d).in_seconds()
        if (dist is None) or (dist_ <= dist):
            min_date = d
            dist = dist_
        else:
            break
    return min_date, dates[_i - 1 :]


def get_files(path: Path) -> Dict[pendulum.DateTime, Dict[str, Path]]:
    """Finds all .xml files in given directory and subdirectories

    Returns a dictionary where the values are dicts of sv & tp identificators with each
    corresponding value being a full path to the file.

    Args:
        path: Path to the root directory to search in

    Returns: Dictionary with the resulting file paths for all dates
    """
    file_d = collections.defaultdict(dict)
    for file in path.glob("**/*.xml"):
        date, file_type = parse_cim_file(file.stem)
        file_d[date][f"{file_type}_path"] = path / file
    return file_d


def get_sv_tp(
    dt: pendulum.DateTime,
    root_path: Path = None,
    file_collection: Dict[pendulum.DateTime, Dict[str, Path]] = None,
) -> Tuple[Dict[str, Path], Dict[pendulum.DateTime, Dict[str, Path]]]:
    """For a given DateTime and a path to a directory of sv & tp files or a collection of parsed files
    from :func:`~cimsparql.parse_xml.get_files` finds and returns the sv/tp pair that are closest to
    the given date.

    Args:
        dt: DateTime in question
        root_path: Path to the root directory to search in
        file_collection:

    Returns: A tuple of a dict of the closest sv/tp pair and a collection of dates not yet parsed.

    """
    file_collection = get_files(root_path) if file_collection is None else file_collection
    min_f, rest_dates = find_min(dt, sorted(list(file_collection.keys())))
    sv_tp = file_collection[min_f]
    file_collection = {k: file_collection[k] for k in rest_dates}
    return sv_tp, file_collection


def get_cim_files(
    root_path: Path, date_range: Iterable[pendulum.DateTime]
) -> Dict[pendulum.DateTime, Dict[str, Path]]:
    """For a given directory and a list/range of DateTimes, finds the paths to the files that are
    closest wrt. to the dates in date_range. Assumes date_range is sorted in ascending order

    Args:
        root_path: Path to the root folder of a directory with sv & tp files
        date_range: Iterable of pendulum DateRange

    Returns: dictionary with date as key and another dictionary with path to the sv/tp files

    """
    file_collection = None
    results = {}
    for date in date_range:
        sv_tp, file_collection = get_sv_tp(date, root_path, file_collection)
        results[date] = sv_tp
    return results



from enum import auto

from strenum import StrEnum


class Impedance(StrEnum):
    r = auto()
    x = auto()


class GeneratorTypes(StrEnum):
    Hydro = auto()
    Thermal = auto()
    Wind = auto()


class ConverterTypes(StrEnum):
    VoltageSourceConverter = "ALG:VoltageSourceConverter"
    DCConverter = "ALG:DCConverter"
    VsConverter = "cim:VsConverter"
    CsConverter = "cim:CsConverter"
    DCConverterUnit = "cim:DCConvertUnit"


class LoadTypes(StrEnum):
    ConformLoad = auto()
    NonConformLoad = auto()
    EnergyConsumer = auto()


class Power(StrEnum):
    p = auto()
    q = auto()


class Rates(StrEnum):
    Normal = auto()
    Warning = auto()
    Overload = auto()


class TapChangerObjects(StrEnum):
    high = auto()
    low = auto()
    neutral = auto()


class SyncVars(StrEnum):
    sn = auto()
    p = auto()
    q = auto()



from typing import Iterable

from cimsparql.query_support import combine_statements, group_query


def _query_str(var_list: Iterable[str], rdf_type: str, connection: str) -> str:
    select = "SELECT ?mrid " + " ".join([f"?{x}" for x in var_list])
    where = [
        f"?s rdf:type cim:{rdf_type}",
        f"?s cim:{rdf_type}.{connection} ?mrid",
        *[f"?s cim:{rdf_type}.{x} ?{x}" for x in var_list],
    ]
    return combine_statements(select, group_query(where))


def powerflow(power: Iterable[str] = ("p", "q")) -> str:
    return _query_str(power, "SvPowerFlow", "Terminal")


def voltage(voltage_vars: Iterable[str] = ("v", "angle")) -> str:
    return _query_str(voltage_vars, "SvVoltage", "TopologicalNode")


def tapstep() -> str:
    return """
    SELECT ?mrid ?position
    WHERE {
    ?t_mrid rdf:type cim:SvTapStep .
    ?t_mrid cim:SvTapStep.TapChanger ?mrid .
    ?t_mrid cim:SvTapStep.position ?position .
    }
    """



from typing import Dict, Iterable, List, Literal, Optional, Union

from cimsparql.cim import (
    BIDDINGAREA,
    CNODE_CONTAINER,
    DELIVERYPOINT,
    EQUIP_CONTAINER,
    GEO_REG,
    ID_OBJ,
    MARKETCODE,
    OPERATIONAL_LIMIT_SET,
    SUBSTATION,
    TC_EQUIPMENT,
    TC_NODE,
)
from cimsparql.constants import con_mrid_str, sequence_numbers
from cimsparql.typehints import Region


def graph(url: Optional[str], query: str) -> str:
    return f"graph <{url}> {{{query}}}" if url else query


def base_voltage(mrid: str, var: str) -> str:
    return f"{mrid} cim:ConductingEquipment.BaseVoltage/cim:BaseVoltage.nominalVoltage {var}"


def node_list(node: str, query_list: List[str], cim_version: int, mrid: str) -> None:
    query_list.extend(
        [
            f"{mrid} cim:{acdc_terminal(cim_version)}.connected 'true'",
            f"{mrid} cim:Terminal.TopologicalNode {node}",
        ]
    )


def terminal_sequence_query(
    cim_version: int, con: Optional[str], nodes: Optional[str], mrid_subject: str
) -> List[str]:
    query_list = []
    for nr in sequence_numbers:
        t_sequence_mrid = f"?_t_mrid_{nr}"
        query_list.extend(
            [
                f"{t_sequence_mrid} {ID_OBJ}.mRID ?t_mrid_{nr}",
                rdf_type_tripler(t_sequence_mrid, "cim:Terminal"),
                f"{t_sequence_mrid} {TC_EQUIPMENT} {mrid_subject}",
                f"{t_sequence_mrid} cim:{acdc_terminal(15)}.sequenceNumber {nr}",
            ]
        )
        if con:
            query_list.append(f"{t_sequence_mrid} {TC_NODE} ?{con}_{nr}")
        if nodes:
            node_list(f"?{nodes}_{nr}", query_list, cim_version, t_sequence_mrid)
            query_list.append(f"{t_sequence_mrid} cim:ACDCTerminal.connected 'true'")
    return query_list


def operational_limit(
    mrid: str,
    rate: str,
    limit_type: Literal["ActivePowerLimit", "CurrentLimit"] = "ActivePowerLimit",
    limit_set: Literal["Terminal", "Equipment"] = "Equipment",
) -> List[str]:
    equip_predicate = f"{OPERATIONAL_LIMIT_SET}/cim:OperationalLimitSet.{limit_set}"
    return [
        f"?p_lim{rate} {equip_predicate} {mrid}",
        rdf_type_tripler(f"?p_lim{rate}", f"cim:{limit_type}"),
        f"?p_lim{rate} {ID_OBJ}.name '{rate}@20'",
        f"?p_lim{rate} cim:{limit_type}.value ?rate{rate}",
    ]


def region_name_query(region: str, sub_region: bool, geographical_region: str) -> str:
    predicate = "SN:IdentifiedObject.shortName" if sub_region else f"{GEO_REG}.Region/{ID_OBJ}.name"
    return f"{geographical_region} {predicate} {region}"


def region_query(region: Region, sub_region: bool, container: str) -> List[str]:
    if region is None:
        return []
    try:
        regions_str = region if isinstance(region, str) else "|".join(region)
        filter = [f"FILTER regex(?area, '{regions_str}')"]
    except TypeError:
        filter = []
    return [
        f"?{container} cim:{container}.Region ?subgeoreg",
        region_name_query("?area", sub_region, "?subgeoreg"),
        *filter,
    ]


def sequence_variables(var: str = con_mrid_str) -> List[str]:
    return [f"?{var}_{nr}" for nr in sequence_numbers]


def _xsd_type(cim: str, var: str) -> str:
    return f"^^<{cim}{var}>"


def acdc_terminal(cim_version: int) -> str:
    return "ACDCTerminal" if cim_version > 15 else "Terminal"


def predicate_list(subject: str, predicate: str, properties: Dict[str, str]) -> List[str]:
    return [f"{subject} {predicate}.{property} {object}" for property, object in properties.items()]


def rdf_type_tripler(subject: str, predicate: str) -> str:
    return f"{subject} rdf:type {predicate}"


def include_market(with_market: bool, variables: List[str], where_list: List[str]) -> None:
    if with_market:
        variables.extend(sequence_variables("bidzone"))
        where_list.extend([market_code_query(terminal_nr) for terminal_nr in sequence_numbers])


def market_code_query(nr: Optional[int] = None, substation: Optional[str] = None) -> str:
    nr_s = "" if nr is None else f"_{nr}"
    bidzone_predicate = f"{DELIVERYPOINT}/{BIDDINGAREA}/{MARKETCODE}"
    if not substation:
        substation = f"?_t_mrid{nr_s}"
        bidzone_predicate = f"{TC_NODE}/{CNODE_CONTAINER}/{SUBSTATION}/{bidzone_predicate}"
    return f"optional {{{substation} {bidzone_predicate} ?bidzone{nr_s}}}"


def terminal_where_query(
    cim_version: int,
    con: Optional[str],
    node: Optional[str],
    mrid_subject: str,
    with_sequence_number: bool = False,
) -> List[str]:
    t_mrid_subject: str = "?_t_mrid"
    query_list = [
        rdf_type_tripler(t_mrid_subject, "cim:Terminal"),
        f"{t_mrid_subject} {TC_EQUIPMENT} {mrid_subject}",
    ]
    if con:
        query_list.append(f"{t_mrid_subject} {TC_NODE} ?{con}")
    if node:
        query_list.extend(
            [
                f"{t_mrid_subject} cim:{acdc_terminal(cim_version)}.connected 'true'",
                f"{t_mrid_subject} cim:Terminal.TopologicalNode ?{node}",
            ]
        )
    if with_sequence_number:
        query_list.append(
            f"{t_mrid_subject} cim:{acdc_terminal(cim_version)}.sequenceNumber ?sequenceNumber"
        )
    return query_list


def _temperature_list(temperature: float, xsd: str, curve: str) -> List[str]:
    signed_temperature = f"{negpos(temperature)}_{abs(temperature)}"
    subject = f"?t{signed_temperature}"
    percent = f"?{signed_temperature}_factor"
    temperature_value = f"'{temperature:0.1f}'{xsd}"
    properties = {"Curve": curve, "temperature": temperature_value, "percent": percent}
    return predicate_list(subject, "ALG:TemperatureCurveData", properties)


def temp_correction_factors(
    mrid: str, cim: str, temperatures: Iterable, temperature_mrid: str = "?temp_mrid"
) -> List[str]:
    where_list = [
        rdf_type_tripler(temperature_mrid, "ALG:TemperatureCurveDependentLimit"),
        f"{temperature_mrid} ALG:LimitDependency.Equipment {mrid}",
        f"{temperature_mrid} ALG:TemperatureCurveDependentLimit.TemperatureCurve ?tcur",
    ]
    xsd = _xsd_type(cim, "Temperature")
    for temperature in temperatures:
        where_list.extend(_temperature_list(temperature, xsd, "?tcur"))
    return where_list


def bid_market_code_query(mrid_subject: str) -> List[str]:
    return [
        f"{mrid_subject} {EQUIP_CONTAINER} ?eq_container",
        f"?eq_container {SUBSTATION} ?substation",
        f"?substation {DELIVERYPOINT}/{BIDDINGAREA}/{MARKETCODE} ?bidzone",
    ]


def to_variables(vars: Iterable[str]) -> List[str]:
    return [f"?{var}" for var in vars]


def combine_statements(*args, group: bool = False, split: str = "\n") -> str:
    """Join *args

    Args:
       group: return enclosed by {...}
       split: join *args by this

    Example:
       >>> import os
       >>> where_list = ['?mrid rdf:type cim:ACLineSegment', '?mrid cim:ACLineSegment.r ?r']
       >>> combine_statements(where_list,group=True, split='\n')
    """
    return "{\n" + split.join(args) + "\n}" if group else split.join(args)


def negpos(val: Union[float, int]) -> str:
    """Convert 'sign' to text"""
    return "minus" if val < 0 else "plus"


def select_statement(variables: Optional[List[str]] = None, distinct: bool = False) -> str:
    """Combine variables in an select statement"""
    vars = "*" if variables is None else " ".join(variables)
    return f"SELECT {'distinct' if distinct else ''} {vars}"


def group_query(
    x: List[str], command: str = "WHERE", split: str = ".\n", group: bool = True
) -> str:
    """Group Query

    Args:
       x: List of objects to group
       command: to operate on group
       split, group: (see: combine_statements)

    Example:
       >>> import os
       >>> where_list = ['?mrid rdf:type cim:ACLineSegment', '?mrid cim:ACLineSegment.r ?r']
       >>> group_query(where_list, group=True, split= '.\n')
    """
    return command + " " + combine_statements(*x, group=group, split=split)


def unionize(*args: str, group: bool = True) -> str:
    if group:
        args = tuple(f"{{\n{arg}\n}}" for arg in args)
    return "\nUNION\n".join(args)


def get_name(mrid: str, name: str, alias: bool = False) -> str:
    param = "aliasName" if alias else "name"
    return f"{mrid} {ID_OBJ}.{param} {name}"


def terminal_number(
    subject: str, predicat: str, number: Union[str, int], union: bool = True
) -> str:
    if union and (isinstance(number, int) or not number.startswith("?")):
        return unionize(f"{subject} {predicat} {number}", f"{subject} {predicat} '{number}'")
    return f"{subject} {predicat} {number}"


def border_filter(region: Union[str, List[str]], area1: str, area2: str) -> List[str]:
    """Border filter where one area is in and the other is out"""

    def _in_first(var1: str, var2: str, regions: Optional[str]) -> List[str]:
        """Return filter for inclusion of first variable and not second"""
        return [f"FILTER (regex({var1}, '{regions}'))", f"FILTER (!regex({var2}, '{regions}'))"]

    regions = "|".join(region) if isinstance(region, list) else region
    return [
        combine_statements(*_in_first(area1, area2, regions)),
        combine_statements(*_in_first(area2, area1, regions)),
    ]



"""Functions used to configure GraphDB client

Will handle authenticated instances of GraphDB where user and password is given in environment
variables ("GRAPHDB_USER" & "GRAPHDB_USER_PASSWD").

"""
import os
import re
from typing import Dict, ItemsView, Iterable, List, Optional

import requests


def service(
    repo: Optional[str] = None,
    server: str = "127.0.0.1:7200",
    protocol: str = "https",
    path: str = "",
) -> str:
    """Returns service url for GraphdDBClient

    Args:
       repo: Repo on server
       server: server ip/name
       protocol: http or https
    """
    url = f"{protocol}://{server}/{path}repositories"
    if repo:
        url += f"/{repo}"
    return url


class GraphDbConfig:
    def __init__(
        self,
        server: str = "127.0.0.1:7200",
        protocol: str = "https",
        auth: requests.auth.AuthBase = None,
    ) -> None:
        """Get repo configuration from GraphDB

        Args:
           server: GraphDB server
           protocol: http or https

        """
        self._service = service(None, server, protocol)
        if auth is None:
            auth = requests.auth.HTTPBasicAuth(
                os.getenv("GRAPHDB_USER"), os.getenv("GRAPHDB_USER_PASSWD")
            )
        try:
            response = requests.get(
                self._service, headers={"Accept": "application/json"}, auth=auth
            )
            response.raise_for_status()
            self._repos = response.json()["results"]["bindings"]
        except (requests.exceptions.RequestException, KeyError):
            self._repos: List[Dict[str, Dict[str, str]]] = []

    @property
    def repos(self) -> List[str]:
        """List of available repos on GraphDB server"""
        if self._repos:
            return [repo["id"]["value"] for repo in self._repos]
        return []


class Prefix:
    def __init__(self) -> None:
        self._prefixes = None

    @property
    def prefixes(self) -> Dict[str, str]:
        if self._prefixes is None:
            self._prefixes = self._get_prefixes()
        return self._prefixes

    def _get_prefixes(self) -> Dict[str, str]:
        prefixes = {}
        auth = requests.auth.HTTPBasicAuth(self.user, self.passwd)
        response = requests.get(self.service + "/namespaces", auth=auth)
        if response.ok:
            for line in response.text.split():
                prefix, uri = line.split(",")
                if prefix != "prefix":
                    prefixes[prefix] = uri.rstrip("#")
        else:
            msg = (
                "Could not fetch namespaces and prefixes from graphdb "
                "Verify that user and password are correctly set in the "
                "GRAPHDB_USER and GRAPHDB_USER_PASSWD environment variable"
            )
            raise RuntimeError(
                f"{msg}\nStatus code: {response.status_code}\nReason: {response.reason}"
            )
        return prefixes

    def in_prefixes(self, variables: Iterable) -> Iterable:
        return {variable for variable in variables if variable.split(":")[0] in self.prefixes}

    def header_str(self, query: str) -> str:
        """Build header string, for sparql queries, with list of prefixes

        The list of available prefixes should be provided by the source (such as GraphDB).

        """
        try:
            return "\n".join(
                [
                    f"PREFIX {name}:<{self.prefixes[name]}#>"
                    for name in set(re.findall(r"(\w+):\w+", query))
                    if name in self.prefixes
                ]
            )
        except AttributeError:
            return ""

    def items(self) -> ItemsView[str, str]:
        """Get an itemsview of prefixes in graphdb instance"""
        return self.prefixes.items()

    @property
    def cim_version(self) -> int:
        """CIM version on server/repo"""
        return int(self.prefixes["cim"].split("CIM-schema-cim")[1])

    @property
    def ns(self) -> Dict[str, str]:
        """Return namespace as dict"""
        return {name: f"{url}#" for name, url in self.items()}

    @property
    def inverse_ns(self) -> Dict[str, str]:
        """Return inverse namespace as dict"""
        return {f"{url}#": name for name, url in self.items()}



from operator import eq, ne
from typing import Callable, Iterable, List, Optional, Union

import cimsparql.query_support as sup
from cimsparql.cim import (
    EQUIP_CONTAINER,
    GEO_REG,
    ID_OBJ,
    SUBSTATION,
    SYNC_MACH,
    TC_EQUIPMENT,
    TC_NODE,
    TR_WINDING,
)
from cimsparql.constants import sequence_numbers, union_split
from cimsparql.enums import (
    ConverterTypes,
    Impedance,
    LoadTypes,
    Power,
    Rates,
    SyncVars,
    TapChangerObjects,
)
from cimsparql.transformer_windings import number_of_windings, terminal, transformer_common
from cimsparql.typehints import Region


def version_date() -> str:
    name: str = "?name"
    variables = ["?mrid", name, "?activationDate"]
    where_list = [
        sup.rdf_type_tripler("?marketDefinitionSet", "SN:MarketDefinitionSet"),
        f"?marketDefinitionSet {ID_OBJ}.mRID ?mrid",
        sup.get_name("?marketDefinitionSet", name),
        "?marketDefinitionSet SN:MarketDefinitionSet.activationDate ?activationDate",
        f"FILTER regex({name}, 'ScheduleResource')",
    ]
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def regions_query() -> str:
    mrid_subject = "?_mrid"
    variables = ["?mrid", "?shortName"]
    region_variable = "?subgeoreg"
    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        sup.rdf_type_tripler(mrid_subject, GEO_REG),
        f"{mrid_subject} SN:IdentifiedObject.shortName ?shortName",
        f"{mrid_subject} {GEO_REG}.Region {region_variable}",
    ]
    names = {mrid_subject: ["?name", "?alias_name"], region_variable: ["?region", "?region_name"]}
    for name_mrid, (name, alias_name) in names.items():
        where_list.append(sup.get_name(name_mrid, name))
        where_list.append(sup.get_name(name_mrid, alias_name, alias=True))
        variables.extend([name, alias_name])
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def phase_tap_changer_query(
    region: Region,
    sub_region: bool,
    with_tap_changer_values: bool,
    impedance: Iterable[Impedance],
    tap_changer_objects: Iterable[TapChangerObjects],
) -> str:
    mrid_subject = "?_mrid"
    variables = ["?mrid", "?w_mrid_1", "?w_mrid_2", "?t_mrid_1", "?t_mrid_2"]
    tap = "?tap"
    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        sup.rdf_type_tripler(mrid_subject, TR_WINDING),
        f"{mrid_subject} cim:TransformerEnd.PhaseTapChanger {tap}",
        f"{mrid_subject} {TR_WINDING}.PowerTransformer ?pt",
    ]

    if with_tap_changer_values:
        variables.extend([tap, "?phase_incr"] + sup.to_variables(tap_changer_objects))
        properties = {f"{obj}Step": f"?{obj}" for obj in tap_changer_objects}
        where_list.extend(
            [
                *sup.predicate_list(tap, "cim:TapChanger", properties),
                "?tap cim:PhaseTapChangerLinear.stepPhaseShiftIncrement ?phase_incr",
            ]
        )

    if impedance:
        variables.extend(sup.to_variables(impedance))
        where_list.extend([f"?w_mrid_1 {TR_WINDING}.{imp} ?{imp}" for imp in impedance])

    if region:
        where_list.extend([f"?pt {EQUIP_CONTAINER} ?Substation"])
        where_list.extend(sup.region_query(region, sub_region, "Substation"))

    for i in sequence_numbers:
        where_list.extend(
            [
                sup.rdf_type_tripler(f"?term_{i}", "cim:Terminal"),
                f"?w_mrid_{i} {TR_WINDING}.PowerTransformer ?pt",
                f"?w_mrid_{i} cim:TransformerEnd.Terminal ?term_{i}",
                f"?term_{i} cim:Terminal.sequenceNumber {i}",
                f"?term_{i} {ID_OBJ}.mRID ?t_mrid_{i}",
            ]
        )

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def connectivity_names(mrid_subject: str, name: str = "?name") -> str:
    variables = [mrid_subject, name]
    where_list = [
        sup.rdf_type_tripler(mrid_subject, "cim:ConnectivityNode"),
        sup.get_name(mrid_subject, name),
    ]
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def full_model() -> str:
    variables = ["?model", "?time", "?profile", "?description", "?version", "?created", "?dependon"]
    where_list = [
        sup.rdf_type_tripler("?model", "md:FullModel"),
        "?model md:Model.profile ?profile",
        "?model md:Model.scenarioTime ?time",
        "?model md:Model.description ?description",
        "?model md:Model.version ?version",
        "?model md:Model.created ?created",
        "?model md:Model.DependentOn ?dependon",
        "?dependon rdf:type md:FullModel",
    ]
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def bus_data(region: Region, sub_region: bool, with_market: bool = True) -> str:
    mrid_subject = "?t_mrid"
    bus_name = "?busname"

    variables = [f"({mrid_subject} as ?mrid)", "?name", bus_name, "?un"]
    where_list = [
        sup.rdf_type_tripler(mrid_subject, "cim:TopologicalNode"),
        sup.get_name(mrid_subject, bus_name),
        f"{mrid_subject} cim:TopologicalNode.BaseVoltage/cim:BaseVoltage.nominalVoltage ?un",
        f"{mrid_subject} cim:TopologicalNode.ConnectivityNodeContainer ?cont",
        f"?cont {ID_OBJ}.aliasName ?name",
        f"?cont {SUBSTATION} ?Substation",
    ]
    if with_market:
        variables.append("?bidzone")
        where_list.append(sup.market_code_query(substation="?Substation"))

    if region:
        where_list.extend(sup.region_query(region, sub_region, "Substation"))

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def three_winding_dummy_bus(region: Region, sub_region: bool) -> str:
    name = "?name"
    variables = ["?mrid", name, f"({name} as ?busname)", "?un"]
    where_list = [
        f"?p_mrid {ID_OBJ}.mRID ?mrid",
        "?w_mrid cim:TransformerEnd.endNumber 1",
        f"?w_mrid {TR_WINDING}.ratedU ?un",
        f"?w_mrid {TR_WINDING}.PowerTransformer ?p_mrid",
        sup.get_name("?p_mrid", name),
        number_of_windings("?p_mrid", 3),
    ]
    if region:
        where_list.append(f"?p_mrid {EQUIP_CONTAINER} ?Substation")
        where_list.extend(sup.region_query(region, sub_region, "Substation"))
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def load_query(
    load_type: Iterable[LoadTypes],
    load_vars: Optional[Iterable[Power]],
    region: Region,
    sub_region: bool,
    connectivity: Optional[str],
    nodes: Optional[str],
    station_group_optional: bool,
    with_sequence_number: bool,
    network_analysis: bool,
    station_group: bool,
    cim_version: int,
    ssh_graph: Optional[str],
) -> str:
    mrid_subject = "?_mrid"

    if not (load_type and set(load_type).issubset(LoadTypes)):
        raise ValueError(f"load_type should be any combination of {set(LoadTypes)}")

    variables = ["?mrid", "?t_mrid" if nodes is None else f"?{nodes}", "?bidzone", "?name"]

    if with_sequence_number:
        variables.append("?sequenceNumber")

    if connectivity:
        variables.append(f"?{connectivity}")

    cim_types = [sup.rdf_type_tripler(mrid_subject, f"cim:{cim_type}") for cim_type in load_type]

    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        f"{mrid_subject} {ID_OBJ}.aliasName ?name",
        f"?_t_mrid {ID_OBJ}.mRID ?t_mrid",
        sup.combine_statements(*cim_types, group=len(cim_types) > 1, split=union_split),
        *sup.terminal_where_query(
            cim_version, connectivity, nodes, mrid_subject, with_sequence_number
        ),
        *sup.bid_market_code_query(mrid_subject),
    ]

    if load_vars:
        variables.extend([f"?{load}" for load in load_vars])
        where_list.append(
            sup.graph(
                ssh_graph,
                sup.group_query(
                    [f"{mrid_subject} cim:EnergyConsumer.{load} ?{load}" for load in load_vars],
                    command="OPTIONAL",
                ),
            )
        )

    if station_group:
        variables.append("?station_group")
        predicate = "/".join(
            [
                "cim:NonConformLoad.LoadGroup",
                "SN:NonConformLoadGroup.ScheduleResource",
                "SN:ScheduleResource.marketCode",
            ]
        )
        station_group_str = f"{mrid_subject} {predicate} ?station_group"
        where_list.append(
            f"optional {{{station_group_str}}}" if station_group_optional else station_group_str
        )

    if network_analysis:
        where_list.append(f"{mrid_subject} SN:Equipment.networkAnalysisEnable {network_analysis}")

    if region:
        where_list.extend(
            [
                f"{mrid_subject} {EQUIP_CONTAINER}/{SUBSTATION} ?Substation",
                *sup.region_query(region, sub_region, "Substation"),
            ]
        )

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def synchronous_machines_query(
    sync_vars: Iterable[SyncVars],
    region: Region,
    sub_region: bool,
    connectivity: Optional[str],
    nodes: Optional[str],
    station_group_optional: bool,
    cim_version: int,
    with_sequence_number: bool,
    network_analysis: bool,
    u_groups: bool,
    ssh_graph: Optional[str],
) -> str:
    mrid_subject = "?_mrid"
    name = "?name"

    variables = [
        "?mrid",
        name,
        "?t_mrid" if nodes is None else f"?{nodes}",
        "?station_group",
        "?market_code",
        "?maxP",
        "?allocationmax",
        "?allocationWeight",
        "?minP",
        "?bidzone",
        *[f"?{var}" for var in sync_vars],
    ]

    if connectivity:
        variables.append(f"?{connectivity}")

    if with_sequence_number:
        variables.append("?sequenceNumber")

    properties = {"sn": "ratedS", "p": "p", "q": "q"}

    def _power(mrid: str, sync_vars: Iterable[str], op: Callable[[str, str], bool]) -> List[str]:
        return [
            f"{mrid} cim:RotatingMachine.{properties[var]} ?{var}"
            for var in sync_vars
            if op(var, "sn")
        ]

    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        f"?_t_mrid {ID_OBJ}.mRID ?t_mrid",
        sup.get_name(mrid_subject, name),
        sup.rdf_type_tripler(mrid_subject, SYNC_MACH),
        *sup.bid_market_code_query(mrid_subject),
        *sup.terminal_where_query(
            cim_version, connectivity, nodes, mrid_subject, with_sequence_number
        ),
        *[
            f"{mrid_subject} {SYNC_MACH}.{lim} ?{lim}"
            for lim in ["maxQ", "minQ"]
            if lim in variables
        ],
        *_power(mrid_subject, sync_vars, eq),
        sup.graph(
            ssh_graph, sup.group_query(_power(mrid_subject, sync_vars, ne), command="OPTIONAL")
        ),
    ]

    if network_analysis:
        where_list.append(f"{mrid_subject} SN:Equipment.networkAnalysisEnable {network_analysis}")

    station_group = [
        f"{mrid_subject} cim:SynchronousMachine.GeneratingUnit ?gu",
        "?gu SN:GeneratingUnit.marketCode ?market_code",
        "?gu cim:GeneratingUnit.maxOperatingP ?maxP",
        "?gu cim:GeneratingUnit.minOperatingP ?minP",
        "?gu SN:GeneratingUnit.groupAllocationMax ?apctmax",
        "bind(xsd:float(str(?apctmax))*xsd:float(str(?maxP)) / 100.0 as ?allocationmax)",
        "?gu SN:GeneratingUnit.groupAllocationWeight ?allocationWeight",
        "?gu SN:GeneratingUnit.ScheduleResource ?ScheduleResource",
        "?ScheduleResource SN:ScheduleResource.marketCode ?station_group",
        sup.get_name("?ScheduleResource", "?st_gr_n", alias=True),
    ]

    if station_group_optional:
        where_list.append(sup.group_query(station_group, command="OPTIONAL"))
    else:
        where_list.extend(station_group)

    if not u_groups:
        where_list.append("FILTER (!bound(?st_gr_n) || (!regex(?st_gr_n, 'U-')))")

    if region:
        where_list.extend(
            [
                f"{mrid_subject} {EQUIP_CONTAINER}/{SUBSTATION} ?Substation",
                *sup.region_query(region, sub_region, "Substation"),
            ]
        )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def wind_generating_unit_query(network_analysis: bool) -> str:
    mrid_subject = "?_mrid"
    name = "?name"
    variables = [
        "?mrid",
        "?station_group",
        "?market_code",
        "?maxP",
        "?allocationmax",
        "?allocationWeight",
        "?minP",
        name,
        "?plant_mrid",
    ]
    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        sup.rdf_type_tripler(mrid_subject, "cim:WindGeneratingUnit"),
        f"{mrid_subject} cim:GeneratingUnit.maxOperatingP ?maxP",
        f"{mrid_subject} SN:GeneratingUnit.marketCode ?market_code",
        f"{mrid_subject} cim:GeneratingUnit.minOperatingP ?minP",
        sup.get_name(mrid_subject, name),
        f"{mrid_subject} SN:WindGeneratingUnit.WindPowerPlant ?plant_mrid",
        f"{mrid_subject} SN:GeneratingUnit.groupAllocationMax ?apctmax",
        "bind(xsd:float(str(?apctmax))*xsd:float(str(?maxP)) / 100.0 as ?allocationmax)",
        f"{mrid_subject} SN:GeneratingUnit.groupAllocationWeight ?allocationWeight",
        f"{mrid_subject} SN:GeneratingUnit.ScheduleResource ?sr",
        "?sr SN:ScheduleResource.marketCode ?station_group",
    ]

    if network_analysis:
        where_list.append(f"{mrid_subject} SN:Equipment.networkAnalysisEnable {network_analysis}")

    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def two_winding_transformer_query(
    region: Union[str, List[str]],
    sub_region: bool,
    rates: Iterable[Rates],
    network_analysis: bool,
    with_market: bool,
    p_mrid: bool,
    nodes: Optional[str],
    with_loss: bool,
    name: str,
    impedance: Iterable[Impedance],
    cim_version: int,
) -> str:
    term = nodes if nodes else "t_mrid"
    variables = [f"?{term}_{nr}" for nr in sequence_numbers]
    where_list = [*terminal("?p_mrid", 1), *terminal("?p_mrid", 2)]
    for nr in sequence_numbers:
        if nodes:
            sup.node_list(f"?{nodes}_{nr}", where_list, cim_version, mrid=f"?_t_mrid_{nr}")
        else:
            where_list.append(f"?_t_mrid_{nr} {ID_OBJ}.mRID ?t_mrid_{nr}")

    transformer_common(
        2,
        p_mrid,
        name,
        impedance,
        variables,
        where_list,
        with_market,
        region,
        sub_region,
        rates,
        network_analysis,
        with_loss,
    )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def three_winding_transformer_query(
    region: Union[str, List[str]],
    sub_region: bool,
    rates: Iterable[Rates],
    network_analysis: bool,
    with_market: bool,
    p_mrid: bool,
    nodes: Optional[str],
    with_loss: bool,
    name: str,
    impedance: Iterable[Impedance],
    cim_version: int,
) -> str:
    term = nodes if nodes else "t_mrid"
    variables = [f"?{term}_1", f"(?p_mrid_object as ?{term}_2)"]
    where_list = [
        *terminal("?p_mrid", 1, lock_end_number=False),
        f"?p_mrid {ID_OBJ}.mRID ?p_mrid_object",
    ]
    if nodes:
        sup.node_list(f"?{nodes}_1", where_list, cim_version, mrid="?_t_mrid_1")
    else:
        where_list.append(f"?_t_mrid_1 {ID_OBJ}.mRID ?t_mrid_1")

    transformer_common(
        3,
        p_mrid,
        name,
        impedance,
        variables,
        where_list,
        with_market,
        region,
        sub_region,
        rates,
        network_analysis,
        with_loss,
    )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def transformer_query(
    region: Union[str, List[str]],
    sub_region: bool,
    connectivity: str,
    rates: Iterable[Rates],
    network_analysis: bool,
    with_market: bool,
    impedance: Iterable[Impedance],
) -> str:
    mrid_subject = "?_p_mrid"
    name = "?name"

    variables = [name, "?p_mrid", "?w_mrid", "?endNumber", "?un", "?t_mrid"]
    variables.extend(sup.to_variables(impedance))
    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?p_mrid",
        f"?_w_mrid {ID_OBJ}.mRID ?w_mrid",
        f"?_t_mrid {ID_OBJ}.mRID ?t_mrid",
        sup.rdf_type_tripler(mrid_subject, "cim:PowerTransformer"),
        f"?_w_mrid {TR_WINDING}.PowerTransformer {mrid_subject}",
        f"?_w_mrid {TR_WINDING}.ratedU ?un",
        "?_w_mrid cim:TransformerEnd.endNumber ?endNumber",
        "?_w_mrid cim:TransformerEnd.Terminal ?_t_mrid",
        f"?_w_mrid {ID_OBJ}.name {name}",
        *sup.predicate_list("?_w_mrid", TR_WINDING, {z: f"?{z}" for z in impedance}),
    ]
    if with_market:
        variables.append("?bidzone")
        where_list.append(sup.market_code_query())

    if network_analysis:
        where_list.append(f"{mrid_subject} SN:Equipment.networkAnalysisEnable {network_analysis}")

    if connectivity:
        variables.append(f"?{connectivity}")
        where_list.append(f"?_t_mrid {TC_NODE} ?{connectivity}")

    if region:
        where_list.append(f"{mrid_subject} {EQUIP_CONTAINER} ?Substation")
        where_list.extend(sup.region_query(region, sub_region, "Substation"))

    if rates:
        where_rate = []
        for rate in rates:
            variables.append(f"?rate{rate}")
            where_rate.extend(sup.operational_limit(mrid_subject, rate))
        where_list.append(sup.group_query(where_rate, command="OPTIONAL"))
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def transformers_connected_to_converter(
    region: Region, sub_region: bool, converter_types: Iterable[ConverterTypes]
) -> str:
    mrid_subject = "?_mrid"
    name = "?name"
    variables = ["?mrid", "?t_mrid", "?converter_mrid", name]
    converters = [sup.rdf_type_tripler("?volt", converter) for converter in converter_types]
    where_list = [
        f"?volt {ID_OBJ}.mRID ?converter_mrid",
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        f"?_t_mrid {ID_OBJ}.mRID ?t_mrid",
        sup.rdf_type_tripler(mrid_subject, "cim:PowerTransformer"),
        sup.get_name(mrid_subject, name, alias=True),
        f"?_t_mrid {TC_EQUIPMENT} {mrid_subject}",
        f"?_t_mrid {TC_NODE}/^{TC_NODE}/{TC_EQUIPMENT} ?volt",
        sup.combine_statements(*converters, group=len(converters) > 1, split=union_split),
    ]
    if region:
        where_list.append(f"{mrid_subject} {EQUIP_CONTAINER} ?Substation")
        where_list.extend(sup.region_query(region, sub_region, "Substation"))
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def converters(
    region: Region,
    sub_region: bool,
    converter_types: Iterable[ConverterTypes],
    nodes: Optional[str],
    sequence_numbers: Optional[List[int]],
) -> str:
    mrid_subject = "?_mrid"
    name = "?name"

    variables = ["?mrid", name, "?p"]
    converters = [sup.rdf_type_tripler(mrid_subject, converter) for converter in converter_types]
    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        sup.get_name(mrid_subject, name, alias=True),
        sup.combine_statements(*converters, group=len(converters) > 1, split=union_split),
        f"Optional{{{mrid_subject} cim:ACDCConverter.p ?p}}",
    ]

    if sequence_numbers:
        for num in sequence_numbers:
            if nodes:
                node = f"?{nodes}" + (f"_{num}" if len(sequence_numbers) > 1 else "")
                sup.node_list(node, where_list, cim_version=16, mrid=f"?_t_mrid_{num}")
                variables.append(node)
            else:
                variables.append(f"?t_mrid_{num}")
            where_list.extend(
                [
                    f"?_t_mrid_{num} {TC_EQUIPMENT} {mrid_subject}",
                    f"?_t_mrid_{num} cim:Terminal.sequenceNumber {num}",
                    f"?_t_mrid_{num} {ID_OBJ}.mRID ?t_mrid_{num}",
                ]
            )

    if region:
        container = "Substation"
        vc = f"{mrid_subject} {EQUIP_CONTAINER}/{SUBSTATION} ?{container}."
        dc = f"{mrid_subject} ALG:DCConverter.DCPole/{EQUIP_CONTAINER} ?{container}."
        where_list.extend(
            [
                sup.combine_statements(vc, dc, group=True, split=union_split),
                *sup.region_query(region, sub_region, container),
            ]
        )
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))


def connection_query(
    cim_version: int,
    rdf_types: Union[str, Iterable[str]],
    region: Region,
    sub_region: bool,
    connectivity: Optional[str],
    nodes: Optional[str],
) -> str:
    mrid_subject = "?_mrid"

    variables = ["?mrid", *sup.sequence_variables("t_mrid")]

    if connectivity:
        variables.extend(sup.sequence_variables(connectivity))

    rdf_types = [rdf_types] if isinstance(rdf_types, str) else rdf_types
    cim_types = [sup.rdf_type_tripler(mrid_subject, rdf_type) for rdf_type in rdf_types]

    where_list = [
        f"{mrid_subject} {ID_OBJ}.mRID ?mrid",
        sup.combine_statements(*cim_types, group=len(cim_types) > 1, split=union_split),
        *sup.terminal_sequence_query(cim_version, connectivity, nodes, mrid_subject),
    ]

    if region:
        predicate = f"{EQUIP_CONTAINER}/cim:Bay.VoltageLevel/{SUBSTATION}"
        where_list.append(f"{mrid_subject} {predicate} ?Substation")
        where_list.extend(sup.region_query(region, sub_region, "Substation"))
    return sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))



# Miscellaneous cim classes

ACLINE = "cim:ACLineSegment"
BIDDINGAREA = "SN:MarketDeliveryPoint.BiddingArea"
CNODE_CONTAINER = "cim:ConnectivityNode.ConnectivityNodeContainer"
DELIVERYPOINT = "SN:Substation.MarketDeliveryPoint"
EQUIP_CONTAINER = "cim:Equipment.EquipmentContainer"
GEO_REG = "cim:SubGeographicalRegion"
ID_OBJ = "cim:IdentifiedObject"
MARKETCODE = "SN:BiddingArea.marketCode"
OPERATIONAL_LIMIT_SET = "cim:OperationalLimit.OperationalLimitSet"
SUBSTATION = "cim:VoltageLevel.Substation"
SYNC_MACH = "cim:SynchronousMachine"
TC_EQUIPMENT = "cim:Terminal.ConductingEquipment"
TC_NODE = "cim:Terminal.ConnectivityNode"
TR_WINDING = "cim:PowerTransformerEnd"



from cimsparql.query_support import acdc_terminal, combine_statements, group_query


def terminal(cim_version: int) -> str:
    select = "SELECT ?mrid ?connected ?tp_node"
    where = [
        "?mrid rdf:type cim:Terminal",
        "?mrid cim:Terminal.TopologicalNode ?tp_node",
        f"?mrid cim:{acdc_terminal(cim_version)}.connected ?connected",
    ]
    return combine_statements(select, group_query(where))


def topological_node() -> str:
    select = "SELECT ?mrid ?name ?ConnectivityNodeContainer ?BaseVoltage"
    where = [
        "?mrid rdf:type cim:TopologicalNode",
        "?mrid cim:IdentifiedObject.name ?name",
        "?mrid cim:TopologicalNode.ConnectivityNodeContainer ?ConnectivityNodeContainer",
        "?mrid cim:TopologicalNode.BaseVoltage ?base_voltage_mrid",
        "?base_voltage_mrid cim:BaseVoltage.nominalVoltage ?BaseVoltage",
    ]
    return combine_statements(select, group_query(where))



from typing import Iterable, List, Union

import cimsparql.query_support as sup
from cimsparql.cim import (
    BIDDINGAREA,
    DELIVERYPOINT,
    EQUIP_CONTAINER,
    ID_OBJ,
    MARKETCODE,
    TR_WINDING,
)


def _end_number(nr: int, lock_end_number: bool) -> List[str]:
    """Lock winding mrid to specific end number

    Args:
        nr: Lock to this end number
        lock_end_number: return empty list if false
    """
    return [f"?w_mrid_{nr} cim:TransformerEnd.endNumber {nr}"] if lock_end_number else []


def terminal(mrid: str, nr: int, lock_end_number: bool = True) -> List[str]:
    """Where statements for transformer terminals"""
    return [
        *_end_number(nr, lock_end_number),
        f"?w_mrid_{nr} cim:TransformerEnd.Terminal ?_t_mrid_{nr}",
        f"?w_mrid_{nr} {TR_WINDING}.PowerTransformer {mrid}",
    ]


def number_of_windings(mrid: str, winding_count: int, with_loss: bool = False) -> str:
    """Where statement to lock query to transformers with specific number of windings"""
    variables = [mrid, "(count(distinct ?nr) as ?winding_count)"]
    where_list = [
        f"{mrid} rdf:type cim:PowerTransformer",
        f"?wwmrid {TR_WINDING}.PowerTransformer {mrid}",
        "?wwmrid cim:TransformerEnd.endNumber ?nr",
    ]
    if with_loss:
        variables.append("(sum(xsd:float(?sv_p)) as ?pl)")
        where_list.extend(
            [
                "?wwmrid cim:TransformerEnd.Terminal ?p_t_mrid",
                "?sv_t cim:SvPowerFlow.Terminal ?p_t_mrid",
                "?sv_t cim:SvPowerFlow.p ?sv_p",
            ]
        )
    select = sup.combine_statements(sup.select_statement(variables), sup.group_query(where_list))
    return sup.group_query(
        [select, f"group by {mrid}", f"having (?winding_count = {winding_count})"],
        command="",
        split="\n",
    )


def _market(variables: List[str], where_list: List[str], with_market: bool) -> None:
    if with_market:
        variables.extend(["?bidzone_1", "?bidzone_2"])
        where_list.extend(
            [
                f"optional {{?Substation {DELIVERYPOINT}/{BIDDINGAREA}/{MARKETCODE} ?bidzone}}",
                "bind(?bidzone as ?bidzone_1)",
                "bind(?bidzone as ?bidzone_2)",
            ]
        )


def transformer_common(
    winding_count: int,
    p_mrid: bool,
    name: str,
    impedance: Iterable[str],
    variables: List[str],
    where_list: List[str],
    with_market: bool,
    region: Union[str, List[str]],
    sub_region: bool,
    rates: Iterable[str],
    network_analysis: bool,
    with_loss: bool,
) -> None:
    variables.extend([name, "?mrid", *sup.to_variables(impedance), "?un"])
    if p_mrid:
        variables.append("?p_mrid")
    if with_loss:
        if winding_count == 2:
            variables.append("(?pl / 2 as ?pl_1) (?pl / 2 as ?pl_2)")
        elif winding_count == 3:
            variables.append("(xsd:float(0.0) as ?pl_1) (?pl as ?pl_2)")
    where_list.extend(
        [
            sup.get_name("?p_mrid", name),
            f"?w_mrid_1 {TR_WINDING}.ratedU ?un",
            f"?w_mrid_1 {ID_OBJ}.mRID ?mrid",
            number_of_windings("?p_mrid", winding_count, with_loss),
        ]
    )
    where_list.extend(sup.predicate_list("?w_mrid_1", TR_WINDING, {z: f"?{z}" for z in impedance}))

    if with_market or region is not None:
        where_list.append(f"?p_mrid {EQUIP_CONTAINER} ?Substation")
        where_list.extend(sup.region_query(region, sub_region, "Substation"))
        _market(variables, where_list, with_market)

    if network_analysis:
        where_list.append(f"?p_mrid SN:Equipment.networkAnalysisEnable {network_analysis}")

    if rates:
        where_rate = []
        for rate in rates:
            variables.append(f"?rate{rate}")
            where_rate.extend(sup.operational_limit("?_t_mrid_1", rate, limit_set="Terminal"))
        where_list.append(sup.group_query(where_rate, command="OPTIONAL"))



"""Graphdb CIM sparql client"""
import contextlib
import os
from typing import Dict, List, Optional, Tuple

import pandas as pd
from SPARQLWrapper import JSON, SPARQLWrapper

from cimsparql import url
from cimsparql.model import CimModel, Model
from cimsparql.type_mapper import TypeMapper


def data_row(cols: List[str], rows: List[Dict[str, str]]) -> Dict[str, str]:
    """Get a sample row for extraction of data types

    Args:
       cols: queried columns (optional might return None)
       rows: 'results'→'bindings' from SPARQLWrapper

    Returns:
       samples result of all columns in query
    """
    full_row = {}
    for row in rows:
        if set(cols).difference(full_row):
            full_row.update(row)
        else:
            break
    return full_row


class GraphDBClientBase(Model):
    def __init__(
        self,
        service: str,
        mapper: Optional[TypeMapper] = None,
        infer: bool = False,
        sameas: bool = True,
    ) -> None:
        """GraphDB client

        Args:
           service: string with url to graphdb repository. See cimsparql.url.service
           mapper: GraphDBClient with the mapper (Default to self).
           infer: deduce further knowledge based on existing RDF data and a formal set of
           sameas: map same concepts from two or more datasets
        """
        super().__init__(mapper)
        self._setup_client(service, infer, sameas)

    def __str__(self) -> str:
        return f"<GraphDBClient object, service: {self.service}>"

    def _setup_client(self, service: Optional[str], infer: bool, sameas: bool) -> None:
        """Setup client for querying

        Args:
           service: string with url to graphdb repository. See cimsparql.url.service
           infer: deduce further knowledge based on existing RDF data and a formal set of
           sameas: map same concepts from two or more datasets
        """
        self.user = os.getenv("GRAPHDB_USER")
        self.passwd = os.getenv("GRAPHDB_USER_PASSWD")
        self.service = service or url.service(repo=os.getenv("GRAPHDB_REPO", "LATEST"))
        self.sparql = SPARQLWrapper(self.service)
        self.sparql.setReturnFormat(JSON)
        self.sparql.setCredentials(self.user, self.passwd)
        self.sparql.addParameter("infer", str(infer))
        self.sparql.addParameter("sameAs", str(sameas))

    @staticmethod
    def value_getter(d: Dict[str, str]) -> str:
        """Get item of 'value' key if present else None"""
        with contextlib.suppress(KeyError):
            return d["value"]

    @staticmethod
    def _col_map(data_row) -> Dict[str, str]:
        return {
            column: data.get("datatype", data.get("type", None))
            for column, data in data_row.items()
        }

    def _exec_query(self, query: str, limit: Optional[int]):
        self.sparql.setQuery(self._query_with_header(query, limit))

        processed_results = self.sparql.queryAndConvert()

        cols = processed_results["head"]["vars"]
        data = processed_results["results"]["bindings"]
        out = [{c: self.value_getter(row.get(c, {})) for c in cols} for row in data]
        return out, data_row(cols, data)

    def exec_query(self, query: str, limit: Optional[int] = None) -> List[Dict[str, str]]:
        out, _ = self._exec_query(query, limit)
        return out

    def _get_table(self, query: str, limit: Optional[int]) -> Tuple[pd.DataFrame, Dict[str, str]]:
        out, data_row = self._exec_query(query, limit)
        return pd.DataFrame(out), data_row


class GraphDBClient(GraphDBClientBase, CimModel):
    pass


def get_graphdb_client(
    server: str,
    graphdb_repo: str,
    graphdb_path: str = "services/pgm/equipment/",
    protocol: str = "https",
) -> GraphDBClient:
    """Get GraphDB client

    Args:
        server: graphdb server
        graphdb_repo: query this repo
        graphdb_path: Prepend with this path when graphdb_repo is LATEST
        protocol: https or http
    """
    graphdb_path = graphdb_path if graphdb_repo == "LATEST" else ""
    return GraphDBClient(url.service(graphdb_repo, server, protocol, graphdb_path))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cimsparql']

package_data = \
{'': ['*']}

install_requires = \
['SPARQLWrapper',
 'StrEnum',
 'networkx',
 'numpy',
 'pandas',
 'requests',
 'tables']

extras_require = \
{'parse_xml': ['defusedxml', 'lxml', 'pendulum']}

setup_kwargs = {
    'name': 'cimsparql',
    'version': '1.10.1',
    'description': 'CIM query utilities',
    'long_description': '[![PyPI version](https://img.shields.io/pypi/v/cimsparql)](https://pypi.org/project/cimsparql/)\n[![Python Versions](https://img.shields.io/pypi/pyversions/cimsparql)](https://pypi.org/project/cimsparql/)\n[![License: MIT](https://img.shields.io/badge/License-MIT-yellow.svg)](https://opensource.org/licenses/MIT)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![](https://github.com/statnett/data_cache/workflows/Tests/badge.svg)](https://github.com/statnett/cimsparql/actions?query=workflow%3ATests)\n[![codecov](https://codecov.io/gh/statnett/cimsparql/branch/master/graph/badge.svg)](https://codecov.io/gh/statnett/cimsparql)\n\n# CIMSPARQL Query CIM data using sparql\n\nThis Python package provides functionality for reading/parsing cim data from\neither xml files or GraphDB into Python memory as pandas dataframes.\n\nThe package provides a set of predefined functions/queries to load CIM data\nsuch generator or branch data, though the user can easiliy extend or define\ntheir own queries.\n\n## Usage\n\n### Load data using predefined functions/queries\n\n```python\n>>> from cimsparql.graphdb import GraphDBClient\n>>> from cimsparql.url import service\n>>> gdbc = GraphDBClient(service(repo=\'<repo>\', server=127.0.0.1:7200))\n>>> ac_lines = gdbc.ac_lines(limit=3)\n>>> print(ac_lines[[\'name\', \'x\', \'r\', \'bch\']])\n         name       x       r       bch\n0  <branch 1>  1.9900  0.8800  0.000010\n1  <branch 2>  1.9900  0.8800  0.000010\n2  <branch 3>  0.3514  0.1733  0.000198\n```\n\nIn the example above the client will query repo "<repo>" in the default server\n[GraphDB](https://graphdb.ontotext.com) for AC line values.\n\n### Inspect/view predefined queries\n\nTo see the actual sparql use the `dry_run` option:\n\n```python\n>>> from cimsparql.queries import ac_line_query\n>>> print(ac_line_query(limit=3, dry_run=True))\n```\n\nThe resulting string contains all the prefix\'s available in the Graphdb repo\nmaking it easier to copy and past to graphdb. Note that the prefixes are *not*\nrequired in the user specified quires described below.\n\nThe `dry_run` option is available for all the predefined queries.\n\n### Load data using user specified queries\n\n```python\n>>> query = \'SELECT ?mrid where { ?mrid rdf:type cim:ACLineSegment } limit 2\'\n>>> query_result = gdbc.get_table(query)\n>>> print(query_result)\n```\n\n### List of available repos at the server\n\n```python\n>>> from cimsparql.url import GraphDbConfig\n>>> print(GraphDbConfig().repos)\n```\n\n### Prefix and namespace\n\nAvailable namespace for current graphdb client (`gdbc` in the examples above),\nwhich can be used in queries (such as `rdf` and `cim`) can by found by\n\n```python\n>>> print(gdbc.ns)\n{\'wgs\': \'http://www.w3.org/2003/01/geo/wgs84_pos#\',\n \'rdf\': \'http://www.w3.org/1999/02/22-rdf-syntax-ns#\',\n \'owl\': \'http://www.w3.org/2002/07/owl#\',\n \'cim\': \'http://iec.ch/TC57/2010/CIM-schema-cim15#\',\n \'gn\': \'http://www.geonames.org/ontology#\',\n \'xsd\': \'http://www.w3.org/2001/XMLSchema#\',\n \'rdfs\': \'http://www.w3.org/2000/01/rdf-schema#\',\n \'SN\': \'http://www.statnett.no/CIM-schema-cim15-extension#\',\n \'ALG\': \'http://www.alstom.com/grid/CIM-schema-cim15-extension#\'}\n```\n',
    'author': 'Statnett Datascience',
    'author_email': 'Datascience.Drift@Statnett.no',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/statnett/cimsparql.git',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'extras_require': extras_require,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)
